"""Backup related CLI commands."""
import logging
from pathlib import Path
from typing import Iterable, Optional, Union

import click
import housekeeper.store.models as hk_models

from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.apps.slurm.slurm_api import SlurmAPI
from cg.constants.constants import DRY_RUN, FlowCellStatus
from cg.constants.housekeeper_tags import SequencingFileTag
from cg.exc import (
    DsmcAlreadyRunningError,
    FlowCellAlreadyBackedUpError,
    FlowCellEncryptionError,
    FlowCellError,
    PdcError,
)
from cg.meta.backup.backup import BackupAPI, SpringBackupAPI
from cg.meta.backup.pdc import PdcAPI
from cg.meta.encryption.encryption import (
    EncryptionAPI,
    FlowCellEncryptionAPI,
    SpringEncryptionAPI,
)
from cg.meta.tar.tar import TarAPI
from cg.models.cg_config import CGConfig
from cg.models.flow_cell.flow_cell import (
    FlowCellDirectoryData,
    get_flow_cells_from_path,
)
from cg.store import Store
from cg.store.models import Flowcell, Sample

LOG = logging.getLogger(__name__)


@click.group()
@click.pass_obj
def backup(context: CGConfig):
    """Backup utilities"""
    pass


@backup.command("flow-cells")
@DRY_RUN
@click.pass_obj
def backup_flow_cells(context: CGConfig, dry_run: bool):
    """Back-up flow cells."""
    pdc_api = context.pdc_api
    pdc_api.dry_run = dry_run
    status_db: Store = context.status_db
    flow_cells: list[FlowCellDirectoryData] = get_flow_cells_from_path(
        flow_cells_dir=Path(context.flow_cells_dir)
    )
    for flow_cell in flow_cells:
        db_flow_cell: Optional[Flowcell] = status_db.get_flow_cell_by_name(
            flow_cell_name=flow_cell.id
        )
        flow_cell_encryption_api = FlowCellEncryptionAPI(
            binary_path=context.encryption.binary_path,
            dry_run=dry_run,
            encryption_dir=Path(context.backup.encryption_directories.current),
            flow_cell=flow_cell,
            pigz_binary_path=context.pigz.binary_path,
            slurm_api=SlurmAPI(),
            sbatch_parameter=context.backup.slurm_flow_cell_encryption.dict(),
            tar_api=TarAPI(binary_path=context.tar.binary_path, dry_run=dry_run),
        )
        try:
            pdc_api.start_flow_cell_backup(
                db_flow_cell=db_flow_cell,
                flow_cell_encryption_api=flow_cell_encryption_api,
                status_db=status_db,
            )
        except (
            DsmcAlreadyRunningError,
            FlowCellAlreadyBackedUpError,
            FlowCellEncryptionError,
            PdcError,
        ) as error:
            logging.error(f"{error}")


@backup.command("encrypt-flow-cells")
@DRY_RUN
@click.pass_obj
def encrypt_flow_cells(context: CGConfig, dry_run: bool):
    """Encrypt flow cells."""
    status_db: Store = context.status_db
    flow_cells: list[FlowCellDirectoryData] = get_flow_cells_from_path(
        flow_cells_dir=Path(context.flow_cells_dir)
    )
    for flow_cell in flow_cells:
        db_flow_cell: Optional[Flowcell] = status_db.get_flow_cell_by_name(
            flow_cell_name=flow_cell.id
        )
        if db_flow_cell and db_flow_cell.has_backup:
            LOG.debug(f"Flow cell: {flow_cell.id} is already backed-up")
            continue
        flow_cell_encryption_api = FlowCellEncryptionAPI(
            binary_path=context.encryption.binary_path,
            dry_run=dry_run,
            encryption_dir=Path(context.backup.encryption_directories.current),
            flow_cell=flow_cell,
            pigz_binary_path=context.pigz.binary_path,
            slurm_api=SlurmAPI(),
            sbatch_parameter=context.backup.slurm_flow_cell_encryption.dict(),
            tar_api=TarAPI(binary_path=context.tar.binary_path, dry_run=dry_run),
        )
        try:
            flow_cell_encryption_api.start_encryption()
        except (FlowCellError, FlowCellEncryptionError) as error:
            logging.error(f"{error}")


@backup.command("fetch-flow-cell")
@click.option("-f", "--flow-cell-id", help="Retrieve a specific flow cell, ex. 'HCK2KDSXX'")
@DRY_RUN
@click.pass_obj
def fetch_flow_cell(context: CGConfig, dry_run: bool, flow_cell_id: Optional[str] = None):
    """Fetch the first flow cell in the requested queue from backup"""

    pdc_api = context.pdc_api
    pdc_api.dry_run = dry_run
    encryption_api = EncryptionAPI(binary_path=context.encryption.binary_path, dry_run=dry_run)
    tar_api = TarAPI(binary_path=context.tar.binary_path, dry_run=dry_run)
    context.meta_apis["backup_api"] = BackupAPI(
        encryption_api=encryption_api,
        encryption_directories=context.backup.encryption_directories,
        status=context.status_db,
        tar_api=tar_api,
        pdc_api=pdc_api,
        flow_cells_dir=context.flow_cells_dir,
        dry_run=dry_run,
    )
    backup_api: BackupAPI = context.meta_apis["backup_api"]

    status_api: Store = context.status_db
    flow_cell: Optional[Flowcell] = (
        status_api.get_flow_cell_by_name(flow_cell_name=flow_cell_id) if flow_cell_id else None
    )

    if not flow_cell and flow_cell_id:
        LOG.error(f"{flow_cell_id}: not found in database")
        raise click.Abort

    if not flow_cell_id:
        LOG.info("Fetching first flow cell in queue")

    retrieval_time: Optional[float] = backup_api.fetch_flow_cell(flow_cell=flow_cell)

    if retrieval_time:
        hours = retrieval_time / 60 / 60
        LOG.info(f"Retrieval time: {hours:.1}h")
        return

    if not dry_run and flow_cell:
        LOG.info(f"{flow_cell}: updating flow cell status to {FlowCellStatus.REQUESTED}")
        flow_cell.status = FlowCellStatus.REQUESTED
        status_api.session.commit()


@backup.command("archive-spring-files")
@DRY_RUN
@click.pass_context
@click.pass_obj
def archive_spring_files(config: CGConfig, context: click.Context, dry_run: bool):
    """Archive spring files to PDC"""
    housekeeper_api: HousekeeperAPI = config.housekeeper_api
    LOG.info("Getting all spring files from Housekeeper.")
    spring_files: Iterable[hk_models.File] = housekeeper_api.files(
        tags=[SequencingFileTag.SPRING]
    ).filter(hk_models.File.path.like(f"%{config.environment}/{config.demultiplex.out_dir}%"))
    for spring_file in spring_files:
        LOG.info("Attempting encryption and PDC archiving for file %s", spring_file.path)
        if Path(spring_file.path).exists():
            context.invoke(archive_spring_file, spring_file_path=spring_file.path, dry_run=dry_run)
        else:
            LOG.warning(
                "Spring file %s found in Housekeeper, but not on disk! Archiving process skipped!",
                spring_file.path,
            )


@backup.command("archive-spring-file")
@click.argument("spring-file-path", type=click.Path(exists=True))
@DRY_RUN
@click.pass_obj
def archive_spring_file(config: CGConfig, spring_file_path: str, dry_run: bool):
    """Archive a spring file to PDC"""
    housekeeper_api: HousekeeperAPI = config.housekeeper_api
    pdc_api: PdcAPI = PdcAPI(binary_path=config.pdc.binary_path, dry_run=dry_run)
    encryption_api: SpringEncryptionAPI = SpringEncryptionAPI(
        binary_path=config.encryption.binary_path,
        dry_run=dry_run,
    )
    spring_backup_api: SpringBackupAPI = SpringBackupAPI(
        encryption_api=encryption_api,
        hk_api=housekeeper_api,
        pdc_api=pdc_api,
        dry_run=dry_run,
    )
    LOG.debug("Start spring encryption/backup")
    spring_backup_api.encrypt_and_archive_spring_file(Path(spring_file_path))


@backup.command("retrieve-spring-files")
@DRY_RUN
@click.option("-s", "--sample-id", "object_type", flag_value="sample", type=str)
@click.option("-c", "--case-id", "object_type", flag_value="case", type=str)
@click.option("-f", "--flow-cell-id", "object_type", flag_value="flow_cell", type=str)
@click.argument("identifier", type=str)
@click.pass_context
@click.pass_obj
def retrieve_spring_files(
    config: CGConfig,
    context: click.Context,
    object_type: str,
    identifier: str,
    dry_run: bool,
):
    """Retrieve all spring files for a given identity"""
    status_api: Store = config.status_db
    housekeeper_api: HousekeeperAPI = config.housekeeper_api

    samples: list[Sample] = _get_samples(status_api, object_type, identifier)

    for sample in samples:
        latest_version: hk_models.Version = housekeeper_api.last_version(bundle=sample.internal_id)
        spring_files: Iterable[hk_models.File] = housekeeper_api.files(
            bundle=sample.internal_id,
            tags=[SequencingFileTag.SPRING],
            version=latest_version.id,
        )
        for spring_file in spring_files:
            context.invoke(retrieve_spring_file, spring_file_path=spring_file.path, dry_run=dry_run)


def _get_samples(status_api: Store, object_type: str, identifier: str) -> list[Sample]:
    """Gets all samples belonging to a sample, case or flow cell id"""
    get_samples = {
        "sample": status_api.sample,
        "case": status_api.get_samples_by_case_id,
        "flow_cell": status_api.get_samples_from_flow_cell,
    }
    samples: Union[Sample, list[Sample]] = get_samples[object_type](identifier)
    return samples if isinstance(samples, list) else [samples]


@backup.command("retrieve-spring-file")
@click.argument("spring-file-path", type=click.Path())
@DRY_RUN
@click.pass_obj
def retrieve_spring_file(config: CGConfig, spring_file_path: str, dry_run: bool):
    """Retrieve a spring file from PDC"""
    LOG.info("Attempting PDC retrieval and decryption file %s", spring_file_path)
    housekeeper_api: HousekeeperAPI = config.housekeeper_api
    pdc_api: PdcAPI = PdcAPI(binary_path=config.pdc.binary_path, dry_run=dry_run)
    encryption_api: SpringEncryptionAPI = SpringEncryptionAPI(
        binary_path=config.encryption.binary_path,
        dry_run=dry_run,
    )
    LOG.debug("Start spring retrieval if not dry run mode=%s", dry_run)
    spring_backup_api: SpringBackupAPI = SpringBackupAPI(
        encryption_api=encryption_api,
        hk_api=housekeeper_api,
        pdc_api=pdc_api,
        dry_run=dry_run,
    )
    spring_backup_api.retrieve_and_decrypt_spring_file(Path(spring_file_path))



import logging
import operator
from datetime import datetime
from pathlib import Path
from typing import Any, Optional

from cg.constants import Pipeline
from cg.constants.constants import FileExtensions, FileFormat, WorkflowManager
from cg.constants.nextflow import NFX_WORK_DIR
from cg.exc import CgError
from cg.io.controller import WriteFile
from cg.io.yaml import write_yaml_nextflow_style
from cg.meta.workflow.analysis import AnalysisAPI
from cg.meta.workflow.nf_handlers import NextflowHandler, NfTowerHandler
from cg.models.cg_config import CGConfig
from cg.models.nf_analysis import FileDeliverable, PipelineDeliverables
from cg.models.rnafusion.rnafusion import CommandArgs
from cg.utils import Process

LOG = logging.getLogger(__name__)


class NfAnalysisAPI(AnalysisAPI):
    """Parent class for handling NF-core analyses."""

    def __init__(self, config: CGConfig, pipeline: Pipeline):
        super().__init__(config=config, pipeline=pipeline)
        self.pipeline: Pipeline = pipeline
        self.root_dir: Optional[str] = None
        self.nfcore_pipeline_path: Optional[str] = None
        self.references: Optional[str] = None
        self.profile: Optional[str] = None
        self.conda_env: Optional[str] = None
        self.conda_binary: Optional[str] = None
        self.tower_binary_path: Optional[str] = None
        self.tower_pipeline: Optional[str] = None
        self.account: Optional[str] = None
        self.email: Optional[str] = None
        self.compute_env: Optional[str] = None
        self.revision: Optional[str] = None
        self.nextflow_binary_path: Optional[str] = None

    @property
    def root(self) -> str:
        return self.root_dir

    @property
    def process(self):
        if not self._process:
            self._process = Process(
                binary=self.tower_binary_path,
            )
        return self._process

    @process.setter
    def process(self, process: Process):
        self._process = process

    def get_profile(self, profile: Optional[str] = None) -> str:
        """Get NF profiles."""
        return profile or self.profile

    def get_workflow_manager(self) -> str:
        """Get workflow manager from Tower."""
        return WorkflowManager.Tower.value

    def get_case_path(self, case_id: str) -> Path:
        """Path to case working directory."""
        return Path(self.root_dir, case_id)

    def get_sample_sheet_path(self, case_id: str) -> Path:
        """Path to sample sheet."""
        return Path(self.get_case_path(case_id), f"{case_id}_samplesheet").with_suffix(
            FileExtensions.CSV
        )

    @staticmethod
    def get_nextflow_config_path(nextflow_config: Optional[str] = None) -> Optional[Path]:
        """Path to Nextflow config file."""
        if nextflow_config:
            return Path(nextflow_config).absolute()

    def get_trailblazer_config_path(self, case_id: str) -> Path:
        """Return the path to a Trailblazer config file containing Tower IDs."""
        return Path(self.root_dir, case_id, "tower_ids").with_suffix(FileExtensions.YAML)

    def get_deliverables_file_path(self, case_id: str) -> Path:
        """Path to deliverables file for a case."""
        return Path(self.get_case_path(case_id), f"{case_id}_deliverables").with_suffix(
            FileExtensions.YAML
        )

    def get_metrics_deliverables_path(self, case_id: str) -> Path:
        """Return a path where the <case>_metrics_deliverables.yaml file should be located."""
        return Path(self.root_dir, case_id, f"{case_id}_metrics_deliverables").with_suffix(
            FileExtensions.YAML
        )

    def get_params_file_path(self, case_id: str, params_file: Optional[Path] = None) -> Path:
        """Return parameters file or a path where the default parameters file for a case id should be located."""
        if params_file:
            return Path(params_file).absolute()
        return Path((self.get_case_path(case_id)), f"{case_id}_params_file").with_suffix(
            FileExtensions.YAML
        )

    def create_case_directory(self, case_id: str, dry_run: bool = False) -> None:
        """Create case directory."""
        if not dry_run:
            Path(self.get_case_path(case_id=case_id)).mkdir(parents=True, exist_ok=True)

    def get_log_path(self, case_id: str, pipeline: str, log: str = None) -> Path:
        """Path to NF log."""
        if log:
            return log
        launch_time: str = datetime.now().strftime("%Y-%m-%d_%H.%M.%S")
        return Path(
            self.get_case_path(case_id),
            f"{case_id}_{pipeline}_nextflow_{launch_time}",
        ).with_suffix(FileExtensions.LOG)

    def get_workdir_path(self, case_id: str, work_dir: Optional[Path] = None) -> Path:
        """Path to NF work directory."""
        if work_dir:
            return work_dir.absolute()
        return Path(self.get_case_path(case_id), NFX_WORK_DIR)

    @staticmethod
    def extract_read_files(
        metadata: list, forward_read: bool = False, reverse_read: bool = False
    ) -> list[str]:
        """Extract a list of fastq file paths for either forward or reverse reads."""
        if forward_read and not reverse_read:
            read_direction = 1
        elif reverse_read and not forward_read:
            read_direction = 2
        else:
            raise ValueError("Either forward or reverse needs to be specified")
        sorted_metadata: list = sorted(metadata, key=operator.itemgetter("path"))
        return [d["path"] for d in sorted_metadata if d["read"] == read_direction]

    def verify_sample_sheet_exists(self, case_id: str, dry_run: bool = False) -> None:
        """Raise an error if sample sheet file is not found."""
        if not dry_run and not Path(self.get_sample_sheet_path(case_id=case_id)).exists():
            raise ValueError(f"No config file found for case {case_id}")

    def verify_deliverables_file_exists(self, case_id: str) -> None:
        """Raise an error if deliverables files file is not found."""
        if not Path(self.get_deliverables_file_path(case_id=case_id)).exists():
            raise CgError(f"No deliverables file found for case {case_id}")

    def write_params_file(self, case_id: str, pipeline_parameters: dict) -> None:
        """Write params-file for analysis."""
        LOG.debug("Writing parameters file")
        write_yaml_nextflow_style(
            content=pipeline_parameters,
            file_path=self.get_params_file_path(case_id=case_id),
        )

    @staticmethod
    def write_sample_sheet(
        content: list[list[Any]],
        file_path: Path,
        header: list[str],
    ) -> None:
        """Write sample sheet CSV file."""
        LOG.debug("Writing sample sheet")
        if header:
            content.insert(0, header)
        WriteFile.write_file_from_content(
            content=content,
            file_format=FileFormat.CSV,
            file_path=file_path,
        )

    @staticmethod
    def write_deliverables_file(
        deliverables_content: dict, file_path: Path, file_format=FileFormat.YAML
    ) -> None:
        """Write deliverables file."""
        WriteFile.write_file_from_content(
            content=deliverables_content, file_format=file_format, file_path=file_path
        )

    def write_trailblazer_config(self, case_id: str, tower_id: str) -> None:
        """Write Tower IDs to a file used as the Trailblazer config."""
        config_path: Path = self.get_trailblazer_config_path(case_id=case_id)
        LOG.info(f"Writing Tower ID to {config_path.as_posix()}")
        WriteFile.write_file_from_content(
            content={case_id: [tower_id]},
            file_format=FileFormat.YAML,
            file_path=config_path,
        )

    def run_analysis(
        self,
        case_id: str,
        command_args: CommandArgs,
        use_nextflow: bool,
        dry_run: bool = False,
    ) -> None:
        """Execute run analysis with given options."""
        if use_nextflow:
            self.process = Process(
                binary=self.nextflow_binary_path,
                environment=self.conda_env,
                conda_binary=self.conda_binary,
                launch_directory=self.get_case_path(case_id=case_id),
            )
            LOG.info("Pipeline will be executed using nextflow")
            parameters: list[str] = NextflowHandler.get_nextflow_run_parameters(
                case_id=case_id,
                pipeline_path=self.nfcore_pipeline_path,
                root_dir=self.root_dir,
                command_args=command_args.dict(),
            )
            self.process.export_variables(
                export=NextflowHandler.get_variables_to_export(),
            )

            command = self.process.get_command(parameters=parameters)
            LOG.info(f"{command}")
            sbatch_number: int = NextflowHandler.execute_head_job(
                case_id=case_id,
                case_directory=self.get_case_path(case_id=case_id),
                slurm_account=self.account,
                email=self.email,
                qos=self.get_slurm_qos_for_case(case_id=case_id),
                commands=command,
                dry_run=dry_run,
            )
            LOG.info(f"Nextflow head job running as job {sbatch_number}")

        else:
            LOG.info("Pipeline will be executed using tower")
            if command_args.resume:
                from_tower_id: int = command_args.id
                if not from_tower_id:
                    from_tower_id: int = NfTowerHandler.get_last_tower_id(
                        case_id=case_id,
                        trailblazer_config=self.get_trailblazer_config_path(case_id=case_id),
                    )
                LOG.info(f"Pipeline will be resumed from run {from_tower_id}.")
                parameters: list[str] = NfTowerHandler.get_tower_relaunch_parameters(
                    from_tower_id=from_tower_id, command_args=command_args.dict()
                )
            else:
                parameters: list[str] = NfTowerHandler.get_tower_launch_parameters(
                    tower_pipeline=self.tower_pipeline,
                    command_args=command_args.dict(),
                )
            self.process.run_command(parameters=parameters, dry_run=dry_run)
            if self.process.stderr:
                LOG.error(self.process.stderr)
            if not dry_run:
                tower_id = NfTowerHandler.get_tower_id(stdout_lines=self.process.stdout_lines())
                self.write_trailblazer_config(case_id=case_id, tower_id=tower_id)
            LOG.info(self.process.stdout)

    @staticmethod
    def get_deliverables_template_content() -> list[dict]:
        """Return deliverables file template content."""
        raise NotImplementedError

    def get_deliverables_for_case(self, case_id: str) -> PipelineDeliverables:
        """Return PipelineDeliverables for a given case."""
        deliverable_template: list[dict] = self.get_deliverables_template_content()
        files: list[FileDeliverable] = []
        for file in deliverable_template:
            for deliverable_field, deliverable_value in file.items():
                if deliverable_value is None:
                    continue
                file[deliverable_field] = file[deliverable_field].replace("CASEID", case_id)
                file[deliverable_field] = file[deliverable_field].replace(
                    "PATHTOCASE", str(self.get_case_path(case_id=case_id))
                )
            files.append(FileDeliverable(**file))
        return PipelineDeliverables(files=files)



"""Delivery report mock classes and methods."""
import logging
from datetime import datetime
from pathlib import Path
from typing import Optional

from housekeeper.store.models import Version

from cg.apps.coverage import ChanjoAPI
from cg.meta.report.mip_dna import MipDNAReportAPI
from cg.meta.workflow.analysis import AnalysisAPI
from cg.meta.workflow.mip_dna import MipDNAAnalysisAPI
from cg.models.cg_config import CGConfig

LOG = logging.getLogger(__name__)


class MockMipDNAAnalysisAPI(MipDNAAnalysisAPI):
    """Mock MipDNAAnalysisAPI for CLI tests."""

    def __init__(self, config: CGConfig):
        super().__init__(config)

    @property
    def root(self):
        return "/root/path"


class MockHousekeeperMipDNAReportAPI(MipDNAReportAPI):
    """Mock ReportAPI for CLI tests overwriting Housekeeper methods."""

    def __init__(self, config: CGConfig, analysis_api: AnalysisAPI):
        super().__init__(config, analysis_api)

    def add_delivery_report_to_hk(
        self, case_id: str, delivery_report_file: Path, version: Version
    ) -> None:
        """Mocked add_delivery_report_to_hk method."""
        LOG.info(
            f"add_delivery_report_to_hk called with the following args:  case={case_id}, delivery_report_file="
            f"{delivery_report_file}, version={version}"
        )

    def get_delivery_report_from_hk(self, case_id: str, version: Version) -> None:
        """Return mocked delivery report path stored in Housekeeper."""
        LOG.info(
            f"get_delivery_report_from_hk called with the following args: case={case_id}, version={version}"
        )
        return None

    def get_scout_uploaded_file_from_hk(self, case_id: str, scout_tag: str) -> str:
        """Return mocked uploaded to Scout file."""
        LOG.info(
            f"get_scout_uploaded_file_from_hk called with the following args: case={case_id}, scout_tag={scout_tag}"
        )
        return f"path/to/{scout_tag}"


class MockMipDNAReportAPI(MockHousekeeperMipDNAReportAPI):
    """Mock ReportAPI for CLI tests."""

    def __init__(self, config: CGConfig, analysis_api: AnalysisAPI):
        super().__init__(config, analysis_api)

    def create_delivery_report(
        self, case_id: str, analysis_date: datetime, force_report: bool
    ) -> None:
        """Mocked create_delivery_report method."""
        LOG.info(
            f"create_delivery_report called with the following args: case={case_id}, analysis_date={analysis_date}, "
            f"force_report={force_report}",
        )

    def create_delivery_report_file(
        self, case_id: str, directory: Path, analysis_date: datetime, force_report: bool
    ) -> Path:
        """Return mocked delivery report file path."""
        LOG.info(
            f"create_delivery_report_file called with the following args: case={case_id}, directory={directory}, "
            f"analysis_date={analysis_date}, force_report={force_report}"
        )
        return directory


class MockChanjo(ChanjoAPI):
    """Chanjo mocked class."""

    def __init__(self):
        mock_config = {"chanjo": {"config_path": "/mock/path", "binary_path": "/mock/path"}}
        super().__init__(mock_config)

    def sample_coverage(self, sample_id: str, panel_genes: list) -> Optional[dict[str, float]]:
        """Return mocked sample dictionary."""
        sample_coverage = None
        if sample_id == "ADM1":
            sample_coverage = {"mean_coverage": 38.342, "mean_completeness": 99.1}
        elif sample_id == "ADM2":
            sample_coverage = {"mean_coverage": 37.342, "mean_completeness": 97.1}
        elif sample_id == "ADM3":
            sample_coverage = {"mean_coverage": 39.342, "mean_completeness": 98.1}
        return sample_coverage



"""Code that handles CLI commands to upload"""
import datetime as dt
import logging
from pathlib import Path

import click

from cg.apps.tb import TrailblazerAPI
from cg.constants import EXIT_FAIL, EXIT_SUCCESS, Pipeline
from cg.constants.constants import DRY_RUN
from cg.constants.delivery import PIPELINE_ANALYSIS_TAG_MAP
from cg.constants.priority import PRIORITY_TO_SLURM_QOS
from cg.constants.tb import AnalysisTypes
from cg.meta.deliver import DeliverAPI
from cg.meta.rsync import RsyncAPI
from cg.store import Store
from cg.store.models import Family

LOG = logging.getLogger(__name__)


@click.command("clinical-delivery")
@click.pass_context
@click.argument("case_id", required=True)
@DRY_RUN
def upload_clinical_delivery(context: click.Context, case_id: str, dry_run: bool):
    """Links the appropriate files for a case, based on the data_delivery, to the customer folder
    and subsequently uses rsync to upload it to caesar."""

    click.echo(click.style("----------------- Clinical-delivery -----------------"))

    case: Family = context.obj.status_db.get_case_by_internal_id(internal_id=case_id)
    delivery_types: set[str] = case.get_delivery_arguments()
    is_sample_delivery: bool
    is_case_delivery: bool
    is_complete_delivery: bool
    job_id: int
    is_sample_delivery, is_case_delivery = DeliverAPI.get_delivery_scope(
        delivery_arguments=delivery_types
    )
    if not delivery_types:
        LOG.info(f"No delivery of files requested for case {case_id}")
        return

    LOG.debug("Delivery types are: %s", delivery_types)
    for delivery_type in delivery_types:
        DeliverAPI(
            store=context.obj.status_db,
            hk_api=context.obj.housekeeper_api,
            case_tags=PIPELINE_ANALYSIS_TAG_MAP[delivery_type]["case_tags"],
            sample_tags=PIPELINE_ANALYSIS_TAG_MAP[delivery_type]["sample_tags"],
            delivery_type=delivery_type,
            project_base_path=Path(context.obj.delivery_path),
        ).deliver_files(case_obj=case)

    rsync_api: RsyncAPI = RsyncAPI(context.obj)
    is_complete_delivery, job_id = rsync_api.slurm_rsync_single_case(
        case=case,
        dry_run=dry_run,
        sample_files_present=is_sample_delivery,
        case_files_present=is_case_delivery,
    )
    RsyncAPI.write_trailblazer_config(
        {"jobs": [str(job_id)]}, config_path=rsync_api.trailblazer_config_path
    )
    analysis_name: str = f"{case_id}_rsync" if is_complete_delivery else f"{case_id}_partial"
    if not dry_run:
        context.obj.trailblazer_api.add_pending_analysis(
            case_id=analysis_name,
            analysis_type=AnalysisTypes.OTHER,
            config_path=rsync_api.trailblazer_config_path.as_posix(),
            out_dir=rsync_api.log_dir.as_posix(),
            slurm_quality_of_service=PRIORITY_TO_SLURM_QOS[case.priority],
            data_analysis=Pipeline.RSYNC,
            ticket=case.latest_ticket,
        )
    LOG.info("Transfer of case %s started with SLURM job id %s", case_id, job_id)


@click.command("all-fastq")
@click.pass_context
@DRY_RUN
def auto_fastq(context: click.Context, dry_run: bool):
    """Starts upload of all not previously uploaded cases with analysis type fastq to
    clinical-delivery."""
    exit_code: int = EXIT_SUCCESS
    status_db: Store = context.obj.status_db
    trailblazer_api: TrailblazerAPI = context.obj.trailblazer_api
    for analysis_obj in status_db.get_analyses_to_upload(pipeline=Pipeline.FASTQ):
        if analysis_obj.family.analyses[0].uploaded_at:
            LOG.debug(
                f"Newer analysis already uploaded for {analysis_obj.family.internal_id}, skipping"
            )
            continue
        if analysis_obj.upload_started_at:
            if trailblazer_api.is_latest_analysis_completed(
                case_id=analysis_obj.family.internal_id
            ):
                LOG.info(
                    f"The upload for {analysis_obj.family.internal_id} is completed, setting uploaded at to {dt.datetime.now()}"
                )
                analysis_obj.uploaded_at = dt.datetime.now()
            else:
                LOG.debug(
                    f"Upload to clinical-delivery for {analysis_obj.family.internal_id} has already started, skipping"
                )
            continue
        case: Family = analysis_obj.family
        LOG.info(f"Uploading family: {case.internal_id}")
        analysis_obj.upload_started_at = dt.datetime.now()
        try:
            context.invoke(upload_clinical_delivery, case_id=case.internal_id, dry_run=dry_run)
        except Exception as error:
            LOG.error(f"Upload of case {case.internal_id} failed")
            LOG.error(error)
            exit_code: int = EXIT_FAIL
            continue

    if not dry_run:
        status_db.session.commit()

    if exit_code == EXIT_FAIL:
        raise click.Abort



"""Fixtures for the workflow tests."""
import datetime
from pathlib import Path

import pytest

from cg.constants.constants import MicrosaltAppTags, MicrosaltQC, Pipeline
from cg.meta.compress.compress import CompressAPI
from cg.meta.workflow.microsalt import MicrosaltAnalysisAPI
from cg.meta.workflow.mip_dna import MipDNAAnalysisAPI
from cg.models.cg_config import CGConfig
from cg.models.compression_data import CompressionData
from cg.store.models import Family, Sample
from tests.cli.workflow.balsamic.conftest import (
    balsamic_housekeeper_dir,
    fastq_file_l_1_r_1,
    fastq_file_l_2_r_1,
    fastq_file_l_2_r_2,
)
from tests.meta.compress.conftest import compress_api, real_crunchy_api
from tests.meta.upload.scout.conftest import another_sample_id
from tests.store_helpers import StoreHelpers


@pytest.fixture(scope="function")
def populated_compress_spring_api(
    compress_api: CompressAPI, only_spring_bundle: dict, helpers
) -> CompressAPI:
    """Populated compress api fixture with only spring compressed fastq."""
    helpers.ensure_hk_bundle(compress_api.hk_api, only_spring_bundle)

    return compress_api


@pytest.fixture(scope="function")
def populated_compress_api_fastq_spring(
    compress_api: CompressAPI, spring_fastq_mix: dict, helpers
) -> CompressAPI:
    """Populated compress api fixture with both spring and fastq."""
    helpers.ensure_hk_bundle(compress_api.hk_api, spring_fastq_mix)

    return compress_api


@pytest.fixture(name="only_spring_bundle")
def only_spring_bundle() -> dict:
    """Return a dictionary with bundle info in the correct format."""
    return {
        "name": "ADM1",
        "created": "2019-12-24",
        "files": [
            {
                "path": "/path/HVCHCCCXY-l4t21_535422_S4_L004.spring",
                "archive": False,
                "tags": ["spring"],
            },
        ],
    }


@pytest.fixture(name="spring_fastq_mix")
def spring_fastq_mix(compression_object: CompressionData) -> dict:
    """Return a dictionary with bundle info including both fastq and spring files."""

    return {
        "name": "ADM1",
        "created": "2019-12-24",
        "files": [
            {
                "path": str(compression_object.spring_path),
                "archive": False,
                "tags": ["spring"],
            },
            {
                "path": str(compression_object.fastq_first),
                "archive": False,
                "tags": ["fastq"],
            },
            {
                "path": str(compression_object.fastq_second),
                "archive": False,
                "tags": ["fastq"],
            },
        ],
    }


@pytest.fixture(name="microsalt_qc_pass_run_dir_path")
def microsalt_qc_pass_run_dir_path(
    microsalt_qc_pass_lims_project: str, microsalt_analysis_dir: Path
) -> Path:
    """Return a microsalt run dir path fixture that passes QC."""
    return Path(microsalt_analysis_dir, microsalt_qc_pass_lims_project)


@pytest.fixture(name="microsalt_qc_fail_run_dir_path")
def microsalt_qc_fail_run_dir_path(
    microsalt_qc_fail_lims_project: str, microsalt_analysis_dir: Path
) -> Path:
    """Return a microsalt run dir path fixture that fails QC."""
    return Path(microsalt_analysis_dir, microsalt_qc_fail_lims_project)


@pytest.fixture(name="microsalt_qc_pass_lims_project")
def microsalt_qc_pass_lims_project() -> str:
    """Return a microsalt LIMS project id that passes QC."""
    return "ACC22222_qc_pass"


@pytest.fixture(name="microsalt_qc_fail_lims_project")
def microsalt_qc_fail_lims_project() -> str:
    """Return a microsalt LIMS project id that fails QC."""
    return "ACC11111_qc_fail"


@pytest.fixture(name="microsalt_case_qc_pass")
def microsalt_case_qc_pass() -> str:
    """Return a microsalt case to pass QC."""
    return "microsalt_case_qc_pass"


@pytest.fixture(name="microsalt_case_qc_fail")
def microsalt_case_qc_fail() -> str:
    """Return a microsalt case to fail QC."""
    return "microsalt_case_qc_fail"


@pytest.fixture(name="qc_pass_microsalt_samples")
def qc_pass_microsalt_samples() -> list[str]:
    """Return a list of 20 microsalt samples internal_ids."""
    return [f"ACC22222A{i}" for i in range(1, 21)]


@pytest.fixture(name="qc_fail_microsalt_samples")
def qc_fail_microsalt_samples() -> list[str]:
    """Return a list of 20 microsalt samples internal_ids."""
    return [f"ACC11111A{i}" for i in range(1, 21)]


@pytest.fixture(name="qc_microsalt_context")
def qc_microsalt_context(
    cg_context: CGConfig,
    helpers: StoreHelpers,
    microsalt_case_qc_pass: str,
    microsalt_case_qc_fail: str,
    qc_pass_microsalt_samples: list[str],
    qc_fail_microsalt_samples: list[str],
    microsalt_qc_pass_lims_project: str,
    microsalt_qc_fail_lims_project: str,
) -> CGConfig:
    """Return a Microsalt CG context."""
    analysis_api = MicrosaltAnalysisAPI(cg_context)
    store = analysis_api.status_db

    # Create MWR microsalt case that passes QC
    microsalt_case_qc_pass: Family = helpers.add_case(
        store=store,
        internal_id=microsalt_case_qc_pass,
        name=microsalt_case_qc_pass,
        data_analysis=Pipeline.MICROSALT,
    )

    for sample in qc_pass_microsalt_samples:
        sample_to_add: Sample = helpers.add_sample(
            store=store,
            internal_id=sample,
            application_tag=MicrosaltAppTags.MWRNXTR003,
            application_type=MicrosaltAppTags.PREP_CATEGORY,
            reads=MicrosaltQC.TARGET_READS,
            reads_updated_at=datetime.datetime.now(),
        )

        helpers.add_relationship(store=store, case=microsalt_case_qc_pass, sample=sample_to_add)

    # Create a microsalt MWX case that fails QC
    microsalt_case_qc_fail: Family = helpers.add_case(
        store=store,
        internal_id=microsalt_case_qc_fail,
        name=microsalt_case_qc_fail,
        data_analysis=Pipeline.MICROSALT,
    )

    for sample in qc_fail_microsalt_samples:
        sample_to_add: Sample = helpers.add_sample(
            store=store,
            internal_id=sample,
            application_tag=MicrosaltAppTags.MWXNXTR003,
            application_type=MicrosaltAppTags.PREP_CATEGORY,
            reads=MicrosaltQC.TARGET_READS,
            reads_updated_at=datetime.datetime.now(),
        )

        helpers.add_relationship(store=store, case=microsalt_case_qc_fail, sample=sample_to_add)

    # Setting the target reads to correspond with statusDB
    store.get_application_by_tag(
        tag=MicrosaltAppTags.MWRNXTR003
    ).target_reads = MicrosaltQC.TARGET_READS
    store.get_application_by_tag(
        tag=MicrosaltAppTags.MWXNXTR003
    ).target_reads = MicrosaltQC.TARGET_READS

    cg_context.meta_apis["analysis_api"] = analysis_api

    return cg_context


@pytest.fixture(name="rnafusion_metrics")
def rnafusion_metrics() -> dict[str, float]:
    """Return Rnafusion raw analysis metrics dictionary."""
    return {
        "after_filtering_gc_content": 0.516984,
        "after_filtering_q20_rate": 0.974834,
        "after_filtering_q30_rate": 0.929476,
        "after_filtering_read1_mean_length": 99.0,
        "before_filtering_total_reads": 149984042.0,
        "bias_5_3": 1.07,
        "pct_adapter": 12.005654574904709,
        "pct_mrna_bases": 85.9731,
        "pct_ribosomal_bases": 0.6581,
        "pct_surviving": 99.42004630065911,
        "pct_duplication": 14.8643,
        "reads_aligned": 72391566.0,
        "uniquely_mapped_percent": 91.02,
    }


@pytest.fixture(name="mip_analysis_api")
def fixture_mip_analysis_api(
    cg_context: CGConfig, mip_hk_store, analysis_store
) -> MipDNAAnalysisAPI:
    """Return a MIP analysis API."""
    analysis_api = MipDNAAnalysisAPI(cg_context)
    analysis_api.housekeeper_api = mip_hk_store
    analysis_api.status_db = analysis_store
    return analysis_api



from datetime import datetime

from cg.store import Store
from cg.store.models import Application, ApplicationVersion


def test_add_application_version(
    store_with_an_application_with_and_without_attributes: Store,
    timestamp: datetime,
    prices: dict[str, int],
    version: int = 1,
):
    """Test that the functions returns an application version."""
    # GIVEN a store with applications but without application versions
    applications: list[
        Application
    ] = store_with_an_application_with_and_without_attributes.get_applications()
    store_with_an_application_with_and_without_attributes._get_query(table=ApplicationVersion).all()
    assert len(applications) > 0
    assert (
        len(
            store_with_an_application_with_and_without_attributes._get_query(
                table=ApplicationVersion
            ).all()
        )
        == 0
    )

    # WHEN adding a new application version
    version: ApplicationVersion = (
        store_with_an_application_with_and_without_attributes.add_application_version(
            application=applications[0], version=version, valid_from=timestamp, prices=prices
        )
    )
    store_with_an_application_with_and_without_attributes.session.add(version)

    # THEN the store has one application version
    assert (
        len(
            store_with_an_application_with_and_without_attributes._get_query(
                table=ApplicationVersion
            ).all()
        )
        == 1
    )



"""Tests for meta compress functionality that updates housekeeper."""
from pathlib import Path
from typing import Generator

from housekeeper.store.models import File, Version

from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.constants import HK_FASTQ_TAGS, SequencingFileTag
from cg.meta.compress import CompressAPI, files
from cg.store import Store
from cg.store.models import Sample
from tests.cli.conftest import MockCompressAPI
from tests.meta.compress.conftest import MockCompressionData
from tests.store_helpers import StoreHelpers


def test_get_flow_cell_id_when_hiseqx(
    compress_api: CompressAPI, bcl2fastq_flow_cell_id: str, bcl2fastq_flow_cell_full_name: str
):
    """Test extracting the flow cell id from a fastq file path."""

    # GIVEN a CompressAPI and a flow cell id within a fastq file path
    fastq_path: Path = compress_api.demux_root.joinpath(
        Path(
            bcl2fastq_flow_cell_full_name,
            f"{bcl2fastq_flow_cell_id}-l6t11_Undetermined_GACGTCTT_L006_R1_001.fastq.gz",
        )
    )

    # WHEN retrieving the flow cell id
    returned_flow_cell_name: str = compress_api.get_flow_cell_id(fastq_path=fastq_path)

    # THEN the flow cell id retrieved should be identical to the flow cell id used
    assert returned_flow_cell_name == bcl2fastq_flow_cell_id


def test_get_flow_cell_id_when_novaseq(
    compress_api: CompressAPI, bcl2fastq_flow_cell_id: str, bcl2fastq_flow_cell_full_name: str
):
    """Test extracting the flow cell id from a fastq file path."""

    # GIVEN a CompressAPI and a flow cell id within a fastq file path
    fastq_path: Path = compress_api.demux_root.joinpath(
        Path(
            bcl2fastq_flow_cell_full_name,
            f"{bcl2fastq_flow_cell_id}_ACC10950A36_S36_L001_R1_001.fastq.gz",
        )
    )

    # WHEN retrieving the flow cell id
    returned_flow_cell_name: str = compress_api.get_flow_cell_id(fastq_path=fastq_path)

    # THEN the flow cell id retrieved should be identical to the flow cell id used
    assert returned_flow_cell_name == bcl2fastq_flow_cell_id


def test_add_fastq_housekeeper_when_no_fastq_in_hk(
    compress_api: MockCompressAPI,
    real_housekeeper_api: Generator[HousekeeperAPI, None, None],
    decompress_hk_spring_bundle: dict,
    compression_files: MockCompressionData,
    store: Store,
    helpers: StoreHelpers,
):
    """Test adding fastq files to Housekeeper when no fastq files in Housekeeper."""

    # GIVEN real Housekeeper API populated with a bundle with SPRING metadata
    hk_bundle: dict = decompress_hk_spring_bundle
    sample_id: str = hk_bundle["name"]
    helpers.ensure_hk_bundle(real_housekeeper_api, hk_bundle)
    sample: Sample = helpers.add_sample(store, internal_id=sample_id)
    compress_api.hk_api = real_housekeeper_api

    # GIVEN that there are no FASTQ files in HK
    version: Version = compress_api.hk_api.get_latest_bundle_version(bundle_name=sample_id)
    file_tags: set = set()
    for file_obj in version.files:
        for tag in file_obj.tags:
            file_tags.add(tag.name)
    assert not set(HK_FASTQ_TAGS).intersection(file_tags)

    # WHEN adding the files to housekeeper
    compress_api.add_fastq_hk(
        sample_internal_id=sample.internal_id,
        fastq_first=compression_files.fastq_first_file,
        fastq_second=compression_files.fastq_second_file,
    )

    # THEN assert that the FASTQ files where added to HK
    file_tags = set()
    for file_obj in version.files:
        for tag in file_obj.tags:
            file_tags.add(tag.name)

    assert set(HK_FASTQ_TAGS).intersection(file_tags)


def test_add_decompressed_fastq(
    compress_api: MockCompressAPI,
    real_housekeeper_api: Generator[HousekeeperAPI, None, None],
    decompress_hk_spring_bundle: dict,
    compression_files: MockCompressionData,
    store: Store,
    helpers: StoreHelpers,
    hk_bundle_sample_path: Path,
):
    """Test functionality to add decompressed FASTQ files."""

    # GIVEN real HK API populated with a HK bundle with SPRING info
    hk_bundle: dict = decompress_hk_spring_bundle
    sample_id: str = hk_bundle["name"]
    sample: Sample = helpers.add_sample(store, internal_id=sample_id)
    helpers.ensure_hk_bundle(real_housekeeper_api, hk_bundle)
    compress_api.hk_api = real_housekeeper_api

    # GIVEN that there exists a SPRING archive, spring metadata and unpacked FASTQs
    version: Version = compress_api.hk_api.get_latest_bundle_version(bundle_name=sample_id)
    fastq_first: Path = compression_files.fastq_first_file
    fastq_second: Path = compression_files.fastq_second_file
    spring_file: Path = compression_files.spring_file
    spring_metadata_file: Path = compression_files.updated_spring_metadata_file
    assert files.is_file_in_version(version_obj=version, path=spring_file)
    assert files.is_file_in_version(version_obj=version, path=spring_metadata_file)
    assert not files.is_file_in_version(version_obj=version, path=fastq_first)
    assert not files.is_file_in_version(version_obj=version, path=fastq_second)

    # WHEN adding decompressed files
    was_decompressed: bool = compress_api.add_decompressed_fastq(sample=sample)

    # THEN check that the files were actually decompressed
    assert was_decompressed

    # THEN assert that the FASTQ files have been added with a relative path
    assert files.is_file_in_version(
        version_obj=version, path=Path(hk_bundle_sample_path, fastq_first.name)
    )
    assert files.is_file_in_version(
        version_obj=version, path=Path(hk_bundle_sample_path, fastq_second.name)
    )


def test_get_fastq_tag_names(compress_api: MockCompressAPI, helpers, hk_sample_bundle, sample_id):
    """Tests that we get the correct fastq tags when providing the path to an existing spring file."""

    # GIVEN a spring file
    helpers.ensure_hk_bundle(store=compress_api.hk_api, bundle_data=hk_sample_bundle)
    spring_file: File = compress_api.hk_api.files(tags=[SequencingFileTag.SPRING]).first()
    assert spring_file

    # GIVEN that the spring_file is tagged with more than 'spring'
    assert len(spring_file.tags) > 1 and SequencingFileTag.SPRING in [
        tag.name for tag in spring_file.tags
    ]

    # WHEN getting fastq tags from its path
    fastq_tags: list[str] = compress_api.get_fastq_tag_names(sample_id)

    # THEN the fastq tags should contain all non-spring tags of the spring file
    for tag_name in [tag.name for tag in spring_file.tags]:
        if tag_name != SequencingFileTag.SPRING:
            assert tag_name in fastq_tags

    # THEN the fastq tags should contain the fastq tag
    assert SequencingFileTag.FASTQ in fastq_tags

    # THEN the fastq tags should not contain the spring tag
    assert SequencingFileTag.SPRING not in fastq_tags



from enum import Enum
from typing import Callable, Optional

from sqlalchemy.orm import Query

from cg.store.models import Organism


def filter_organism_by_internal_id(organisms: Query, internal_id: str, **kwargs) -> Query:
    """Return organism by internal id."""
    return organisms.filter(Organism.internal_id == internal_id)


class OrganismFilter(Enum):
    """Define Organism filter functions."""

    FILTER_BY_INTERNAL_ID: Callable = filter_organism_by_internal_id


def apply_organism_filter(
    organisms: Query,
    filter_functions: list[Callable],
    internal_id: Optional[str] = None,
) -> Query:
    """Apply filtering functions and return filtered results."""
    for filter_function in filter_functions:
        organisms: Query = filter_function(
            organisms=organisms,
            internal_id=internal_id,
        )
    return organisms



"""Test CG CLI upload observations module."""

import logging

import pytest
from _pytest.logging import LogCaptureFixture
from click.testing import CliRunner

from cg.cli.upload.observations import upload_observations_to_loqusdb
from cg.cli.upload.observations.utils import (
    get_observations_api,
    get_observations_case,
    get_observations_case_to_upload,
    get_sequencing_method,
)
from cg.constants import EXIT_SUCCESS
from cg.constants.constants import Pipeline
from cg.constants.sequencing import SequencingMethod
from cg.constants.subject import PhenotypeStatus
from cg.exc import CaseNotFoundError, LoqusdbUploadCaseError
from cg.meta.observations.mip_dna_observations_api import MipDNAObservationsAPI
from cg.models.cg_config import CGConfig
from cg.store import Store
from cg.store.models import Family, Sample
from tests.store_helpers import StoreHelpers


def test_observations(
    base_context: CGConfig, cli_runner: CliRunner, helpers: StoreHelpers, caplog: LogCaptureFixture
):
    """Test upload of observations."""
    caplog.set_level(logging.DEBUG)
    store: Store = base_context.status_db

    # GIVEN a case ready to be uploaded to Loqusdb
    case: Family = helpers.add_case(store)
    case.customer.loqus_upload = True
    sample: Sample = helpers.add_sample(store, application_type=SequencingMethod.WES)
    link = store.relate_sample(family=case, sample=sample, status=PhenotypeStatus.UNKNOWN)
    store.session.add(link)

    # WHEN trying to do a dry run upload to Loqusdb
    result = cli_runner.invoke(
        upload_observations_to_loqusdb, [case.internal_id, "--dry-run"], obj=base_context
    )

    # THEN the execution should have been successful and stop at a dry run step
    assert result.exit_code == EXIT_SUCCESS
    assert "Dry run" in caplog.text


def test_get_observations_case(base_context: CGConfig, helpers: StoreHelpers):
    """Test get observations supported case."""
    store: Store = base_context.status_db

    # GIVEN an observations valid case
    case: Family = helpers.add_case(store)

    # WHEN retrieving a case given a specific case ID
    extracted_case = get_observations_case(base_context, case.internal_id, upload=True)

    # THEN the extracted case should match the stored one
    assert extracted_case == case


def test_get_observations_case_invalid_id(base_context: CGConfig, caplog: LogCaptureFixture):
    """Test get observations case providing an incorrect ID."""
    caplog.set_level(logging.DEBUG)

    # WHEN retrieving a case given a specific case ID
    with pytest.raises(CaseNotFoundError):
        # THEN a CaseNotFoundError should be raised
        get_observations_case(base_context, "invalid_case_id", upload=True)

    assert "Invalid case ID. Retrieving available cases for Loqusdb actions." in caplog.text


def test_get_observations_case_to_upload(base_context: CGConfig, helpers: StoreHelpers):
    """Test get case ready to be uploaded to Loqusdb."""
    store: Store = base_context.status_db

    # GIVEN a case ready to be uploaded to Loqusdb
    case: Family = helpers.add_case(store)
    case.customer.loqus_upload = True

    # WHEN retrieving a case given a specific case ID
    extracted_case = get_observations_case_to_upload(base_context, case.internal_id)

    # THEN the extracted case should match the stored one
    assert extracted_case == case


def test_get_observations_api(base_context: CGConfig, helpers: StoreHelpers):
    """Test get observation API given a Loqusdb supported case."""
    store: Store = base_context.status_db

    # GIVEN a Loqusdb supported case
    case: Family = helpers.add_case(store, data_analysis=Pipeline.MIP_DNA)
    sample: Sample = helpers.add_sample(store, application_type=SequencingMethod.WES)
    link = store.relate_sample(family=case, sample=sample, status=PhenotypeStatus.UNKNOWN)
    store.session.add(link)

    # WHEN retrieving the observation API
    observations_api: MipDNAObservationsAPI = get_observations_api(base_context, case)

    # THEN a MIP-DNA API should be returned
    assert observations_api
    assert isinstance(observations_api, MipDNAObservationsAPI)


def test_get_sequencing_method(base_context: CGConfig, helpers: StoreHelpers):
    """Test sequencing method extraction for Loqusdb upload."""
    store: Store = base_context.status_db

    # GIVEN a case object with a WGS sequencing method
    case: Family = helpers.add_case(store)
    sample: Sample = helpers.add_sample(store, application_type=SequencingMethod.WGS)
    link = store.relate_sample(family=case, sample=sample, status=PhenotypeStatus.UNKNOWN)
    store.session.add(link)

    # WHEN getting the sequencing method
    sequencing_method: SequencingMethod = get_sequencing_method(case)

    # THEN the obtained sequencing method should be WGS
    assert sequencing_method == SequencingMethod.WGS


def test_get_sequencing_method_exception(
    base_context: CGConfig,
    helpers: StoreHelpers,
    wgs_application_tag: str,
    external_wes_application_tag: str,
    caplog: LogCaptureFixture,
):
    """Test sequencing method extraction for Loqusdb upload when a case contains multiple sequencing types."""
    store: Store = base_context.status_db

    # GIVEN a case object with a WGS and WES mixed sequencing methods
    case: Family = helpers.add_case(store)
    sample_wgs: Sample = helpers.add_sample(
        store, application_tag=wgs_application_tag, application_type=SequencingMethod.WGS
    )
    sample_wes: Sample = helpers.add_sample(
        store, application_tag=external_wes_application_tag, application_type=SequencingMethod.WES
    )
    link_1 = store.relate_sample(family=case, sample=sample_wgs, status=PhenotypeStatus.UNKNOWN)
    link_2 = store.relate_sample(family=case, sample=sample_wes, status=PhenotypeStatus.UNKNOWN)
    store.session.add_all([link_1, link_2])

    # WHEN getting the sequencing method
    with pytest.raises(LoqusdbUploadCaseError):
        # THEN a LoqusdbUploadCaseError should be raised
        get_sequencing_method(case)

    assert f"Case {case.internal_id} has a mixed analysis type. Cancelling action." in caplog.text



"""Tests for the prepare_fastq_api"""

from cg.apps.crunchy import CrunchyAPI
from cg.meta.compress.compress import CompressAPI
from cg.meta.workflow.prepare_fastq import PrepareFastqAPI
from cg.store import Store
from cg.store.models import Family


def test_is_spring_decompression_needed_when_true(
    populated_compress_spring_api: CompressAPI,
    analysis_store_single_case: Store,
    case_id: str,
    sample_id: str,
):
    """Test when spring decompression is needed."""

    # GIVEN a populated prepare_fastq_api
    prepare_fastq_api = PrepareFastqAPI(
        store=analysis_store_single_case, compress_api=populated_compress_spring_api
    )
    # GIVEN a store with a case that has linked samples
    case_obj: Family = analysis_store_single_case.get_case_by_internal_id(internal_id=case_id)
    assert case_obj
    # GIVEN that the case has linked samples
    link_objects = [link_obj for link_obj in case_obj.links]
    assert link_objects
    # GIVEN a that there exists a version with only spring in housekeeper
    version_object = populated_compress_spring_api.hk_api.get_latest_bundle_version(
        bundle_name=sample_id
    )
    for file in version_object.files:
        assert file.path.endswith(".spring")

    # WHEN checking if spring decompression is needed
    res = prepare_fastq_api.is_spring_decompression_needed(case_id)

    # THEN assert that spring decompression is needed since there are no fastq files
    assert res is True


def test_is_spring_decompression_needed_when_false(
    populated_compress_api_fastq_spring: CompressAPI,
    analysis_store_single_case: Store,
    case_id: str,
):
    """Test when spring decompression is not needed"""

    # GIVEN a populated prepare_fastq_api
    prepare_fastq_api = PrepareFastqAPI(
        store=analysis_store_single_case, compress_api=populated_compress_api_fastq_spring
    )
    # GIVEN a store with a case that has linked samples
    case_obj: Family = analysis_store_single_case.get_case_by_internal_id(internal_id=case_id)
    assert case_obj
    # GIVEN that the case has linked samples
    link_objects = [link_obj for link_obj in case_obj.links]
    assert link_objects

    # WHEN checking if spring decompression is needed
    res = prepare_fastq_api.is_spring_decompression_needed(case_id)

    # THEN assert that spring decompression is not needed since there are fastq files
    assert res is False


def test_at_least_one_sample_be_decompressed(
    populated_compress_spring_api: CompressAPI,
    analysis_store_single_case: Store,
    case_id: str,
    mocker,
):
    # GIVEN spring decompression is possible
    mocker.patch.object(CrunchyAPI, "is_spring_decompression_possible")
    CrunchyAPI.is_spring_decompression_possible.return_value = True

    # GIVEN a populated prepare_fastq_api
    prepare_fastq_api = PrepareFastqAPI(
        store=analysis_store_single_case, compress_api=populated_compress_spring_api
    )

    # WHEN checking if spring decompression is possible
    res = prepare_fastq_api.can_at_least_one_sample_be_decompressed(case_id=case_id)

    # THEN assert that spring decompression is possible
    assert res is True


def test_no_samples_can_be_decompressed(
    populated_compress_spring_api: CompressAPI,
    analysis_store_single_case: Store,
    case_id: str,
    mocker,
):
    # GIVEN spring decompression is not possible
    mocker.patch.object(CrunchyAPI, "is_spring_decompression_possible")
    CrunchyAPI.is_spring_decompression_possible.return_value = False

    # GIVEN a populated prepare_fastq_api
    prepare_fastq_api = PrepareFastqAPI(
        store=analysis_store_single_case, compress_api=populated_compress_spring_api
    )

    # WHEN checking if spring decompression is possible
    res = prepare_fastq_api.can_at_least_one_sample_be_decompressed(case_id=case_id)

    # THEN assert that spring decompression is not possible
    assert res is False


def test_no_fastq_in_housekeeper(
    populated_compress_spring_api: CompressAPI,
    analysis_store_single_case: Store,
    case_id: str,
    sample_id: str,
):
    """Test when FASTQ needs to be added to Housekeeper."""

    # GIVEN a populated prepare_fastq_api
    prepare_fastq_api = PrepareFastqAPI(
        store=analysis_store_single_case, compress_api=populated_compress_spring_api
    )
    # GIVEN a store with a case that has linked samples
    case_obj: Family = analysis_store_single_case.get_case_by_internal_id(internal_id=case_id)
    assert case_obj
    # GIVEN that the case has linked samples
    link_objects = [link_obj for link_obj in case_obj.links]
    assert link_objects
    # GIVEN a that there exists a version with only spring in housekeeper
    version_object = populated_compress_spring_api.hk_api.get_latest_bundle_version(
        bundle_name=sample_id
    )
    for file in version_object.files:
        assert file.path.endswith(".spring")

    # WHEN checking if fastq files need to be added to housekeeper
    res = prepare_fastq_api.add_decompressed_fastq_files_to_housekeeper(case_id)

    # THEN assert that all fastq files are not there
    assert res is None



"""Fixtures for utils"""
from pathlib import Path

import pytest

from cg.utils import Process


@pytest.fixture(scope="function")
def ls_process():
    """
    list files process
    """
    binary = "ls"
    process = Process(binary=binary)
    return process


@pytest.fixture(scope="function")
def stderr_output():
    """
    std err lines
    """
    lines = (
        "2018-11-29 08:41:38 130-229-8-20-dhcp.local "
        "mongo_adapter.client[77135] INFO Connecting to "
        "uri:mongodb://None:None@localhost:27017\n"
        "2018-11-29 08:41:38 130-229-8-20-dhcp.local "
        "mongo_adapter.client[77135] INFO Connection "
        "established\n"
    )
    return lines


@pytest.fixture(scope="session")
def some_file() -> str:
    """Return a file."""
    return "some_file.txt"


@pytest.fixture(scope="session")
def nested_directory_with_file(tmp_path_factory, some_file: str) -> Path:
    """Return a nested directory with a file."""
    directory = tmp_path_factory.mktemp("nested_directory_with_file")
    sub_directory = directory / "sub_directory"
    sub_directory.mkdir()
    file = sub_directory / some_file
    file.touch()
    return directory


@pytest.fixture(scope="session")
def path_with_directories_and_a_file(
    tmp_path_factory, sub_dir_names: list[str], some_file: str
) -> Path:
    """Return a path with directories and a file in it."""
    directory: Path = tmp_path_factory.mktemp("tmp_dir")
    for sub_dir_name in sub_dir_names:
        sub_dir: Path = Path(directory, sub_dir_name)
        sub_dir.mkdir()
    file = Path(directory, some_file)
    file.touch()
    return directory


@pytest.fixture(scope="session")
def sub_dir_names() -> list[str]:
    return ["sub_dir_1", "sub_dir_2", "sub_dir_3"]



import copy

from cg.meta.report.balsamic import BalsamicReportAPI
from cg.models.balsamic.analysis import BalsamicAnalysis
from cg.models.report.metadata import BalsamicTargetedSampleMetadataModel
from cg.store import Store
from cg.store.models import BedVersion, Family, Sample
from tests.store_helpers import StoreHelpers


def test_get_sample_metadata(
    report_api_balsamic: BalsamicReportAPI,
    case_balsamic: Family,
    helpers: StoreHelpers,
    sample_store: Store,
):
    """Test sample metadata extraction."""

    # GIVEN a tumor sample, a bed version, and the latest metadata
    sample: Sample = helpers.add_sample(store=sample_store)
    sample.internal_id = "ACC0000A1"
    sample.reads = 20000000
    bed_name: str = "GIcfDNA"
    bed_version: BedVersion = helpers.ensure_bed_version(store=sample_store, bed_name=bed_name)
    bed_version.filename = "gicfdna_3.1_hg19_design.bed"
    balsamic_metadata: BalsamicAnalysis = report_api_balsamic.analysis_api.get_latest_metadata(
        case_balsamic.internal_id
    )

    # GIVEN the expected output
    expected_metadata: dict[str, str] = {
        "million_read_pairs": "10.0",
        "duplicates": "93.1",
        "mean_insert_size": "178.19",
        "fold_80": "1.16",
        "bait_set": bed_name,
        "bait_set_version": "3.1",
        "median_target_coverage": "5323.0",
        "pct_250x": "N/A",
        "pct_500x": "N/A",
    }

    # WHEN retrieving the sample metadata
    sample_metadata: BalsamicTargetedSampleMetadataModel = report_api_balsamic.get_sample_metadata(
        case=case_balsamic, sample=sample, analysis_metadata=balsamic_metadata
    )

    # THEN check that the sample metadata is correctly retrieved
    assert sample_metadata.model_dump() == expected_metadata


def test_get_variant_callers(report_api_balsamic, case_id):
    """Test variant callers extraction for TN-PANEL analysis."""

    # GIVEN a mock metadata object
    balsamic_metadata = report_api_balsamic.analysis_api.get_latest_metadata(case_id)

    # GIVEN an expected output for tumor normal panel analysis
    expected_callers = [
        "manta (v1.6.0)",
        "cnvkit",
        "vardict (v2019.06.04=pl526_0)",
        "dnascope",
        "tnhaplotyper",
        "manta_germline (v1.6.0)",
        "haplotypecaller",
        "TNscope_umi",
        "delly (v0.8.7)",
    ]

    # WHEN retrieving the analysis variant callers
    variant_callers = report_api_balsamic.get_variant_callers(balsamic_metadata)

    # THEN check that the callers are correctly identified
    assert variant_callers == expected_callers


def test_get_variant_caller_version(report_api_balsamic, case_id):
    """Tests variant caller version extraction."""

    # GIVEN a tool name and a mock variant caller versions dictionary
    var_caller_name = "manta"
    var_caller_versions = report_api_balsamic.analysis_api.get_latest_metadata(
        case_id
    ).config.bioinfo_tools_version

    # GIVEN the tools mock version
    expected_version = "1.6.0"

    # WHEN retrieving the version of a specific variant caller
    version = report_api_balsamic.get_variant_caller_version(var_caller_name, var_caller_versions)

    # THEN verify that the extracted version is correct
    assert version == expected_version


def test_get_report_accreditation(report_api_balsamic, case_id):
    """Tests report accreditation for a specific BALSAMIC analysis."""

    # GIVEN a mock metadata object and an accredited one
    balsamic_metadata = report_api_balsamic.analysis_api.get_latest_metadata(case_id)

    balsamic_accredited_metadata = copy.deepcopy(balsamic_metadata)
    balsamic_accredited_metadata.config.panel.capture_kit = "gmsmyeloid"

    # WHEN performing the accreditation validation
    unaccredited_report = report_api_balsamic.get_report_accreditation(None, balsamic_metadata)
    accredited_report = report_api_balsamic.get_report_accreditation(
        None, balsamic_accredited_metadata
    )

    # THEN verify that only the panel "gmsmyeloid" reports are validated
    assert not unaccredited_report
    assert accredited_report



import logging
from typing import Optional

import click

from cg.cli.set.case import set_case
from cg.constants import CASE_ACTIONS, Priority
from cg.store import Store
from cg.store.models import Family, Sample
from cg.utils.click.EnumChoice import EnumChoice

CONFIRM = "Continue?"

LOG = logging.getLogger(__name__)


def _get_samples_by_identifiers(identifiers: click.Tuple([str, str]), store: Store) -> list[Sample]:
    """Get samples matched by given set of identifiers"""
    identifier_args = dict(identifiers)
    return list(store.get_samples_by_any_id(**identifier_args))


def _get_cases(identifiers: click.Tuple([str, str]), store: Store) -> list[Family]:
    """Get cases that have samples that match identifiers if given"""
    samples_by_id: list[Sample] = _get_samples_by_identifiers(identifiers, store)
    cases: set[Family] = set()
    for sample in samples_by_id:
        for link in sample.links:
            cases.add(link.family)

    return list(cases)


@click.command("cases")
@click.option(
    "--sample-identifier",
    "identifiers",
    nargs=2,
    type=click.Tuple([str, str]),
    multiple=True,
    help="Give an identifier on sample and the value to use it with, e.g. --sample-identifier "
    "name Prov52",
)
@click.option("-a", "--action", type=click.Choice(CASE_ACTIONS), help="update case action")
@click.option("-c", "--customer-id", type=click.STRING, help="update customer")
@click.option("-g", "--panel", "panel_abbreviations", multiple=True, help="update gene panels")
@click.option(
    "-p", "--priority", type=EnumChoice(Priority, use_value=False), help="update priority"
)
@click.pass_context
def set_cases(
    context: click.Context,
    action: Optional[str],
    priority: Optional[Priority],
    panel_abbreviations: Optional[tuple[str]],
    customer_id: Optional[str],
    identifiers: click.Tuple([str, str]),
):
    """Set values on many families at the same time"""
    store: Store = context.obj.status_db
    cases_to_alter: list[Family] = _get_cases(identifiers, store)

    if not cases_to_alter:
        LOG.error("No cases to alter!")
        raise click.Abort

    LOG.info("Would alter cases:")

    for case_to_alter in cases_to_alter:
        LOG.info(case_to_alter)

    if not (click.confirm(CONFIRM)):
        raise click.Abort

    for case_to_alter in cases_to_alter:
        context.invoke(
            set_case,
            action=action,
            priority=priority,
            panel_abbreviations=panel_abbreviations,
            case_id=case_to_alter.internal_id,
            customer_id=customer_id,
        )



"""Utility functions for the upload cli commands."""

import logging
from typing import Optional

import click

from cg.constants import Pipeline
from cg.constants.constants import MAX_ITEMS_TO_RETRIEVE
from cg.store import Store
from cg.store.models import Analysis

LOG = logging.getLogger(__name__)


def suggest_cases_to_upload(status_db: Store, pipeline: Optional[Pipeline] = None) -> None:
    """Print a list of suggested cases to upload."""
    LOG.warning("Provide a case, suggestions:")
    records: list[Analysis] = status_db.get_analyses_to_upload(pipeline=pipeline)[
        :MAX_ITEMS_TO_RETRIEVE
    ]
    for case_obj in records:
        click.echo(case_obj)



"""Class to hold information about scout load config"""

from datetime import datetime
from typing import Optional

from pydantic import BaseModel, BeforeValidator, ConfigDict
from typing_extensions import Annotated, Literal

from cg.models.scout.validators import field_not_none


class ChromographImages(BaseModel):
    autozygous: Optional[str] = None
    coverage: Optional[str] = None
    upd_regions: Optional[str] = None
    upd_sites: Optional[str] = None


class Reviewer(BaseModel):
    alignment: Optional[str] = None
    alignment_index: Optional[str] = None
    vcf: Optional[str] = None
    catalog: Optional[str] = None


class ScoutIndividual(BaseModel):
    alignment_path: Optional[str] = None
    analysis_type: Annotated[
        Literal[
            "external",
            "mixed",
            "panel",
            "panel-umi",
            "unknown",
            "wes",
            "wgs",
            "wts",
        ],
        BeforeValidator(field_not_none),
    ] = None
    capture_kit: Optional[str] = None
    confirmed_parent: Optional[bool] = None
    confirmed_sex: Optional[bool] = None
    father: Optional[str] = None
    mother: Optional[str] = None
    phenotype: Optional[str] = None
    sample_id: Annotated[str, BeforeValidator(field_not_none)] = None
    sample_name: Optional[str] = None
    sex: Annotated[Optional[str], BeforeValidator(field_not_none)] = None
    subject_id: Optional[str] = None
    tissue_type: Optional[str] = None

    model_config = ConfigDict(validate_assignment=True)


class ScoutMipIndividual(ScoutIndividual):
    mt_bam: Optional[str] = None
    chromograph_images: ChromographImages = ChromographImages()
    reviewer: Reviewer = Reviewer()
    rhocall_bed: Optional[str] = None
    rhocall_wig: Optional[str] = None
    tiddit_coverage_wig: Optional[str] = None
    upd_regions_bed: Optional[str] = None
    upd_sites_bed: Optional[str] = None
    vcf2cytosure: Optional[str] = None
    mitodel_file: Optional[str] = None


class ScoutCancerIndividual(ScoutIndividual):
    tumor_type: Optional[str] = None
    tmb: Optional[str] = None
    msi: Optional[str] = None
    tumor_purity: float = 0
    vcf2cytosure: Optional[str] = None


class ScoutLoadConfig(BaseModel):
    owner: Annotated[str, BeforeValidator(field_not_none)] = None
    family: Annotated[str, BeforeValidator(field_not_none)] = None
    family_name: Optional[str] = None
    synopsis: Optional[str] = None
    phenotype_terms: Optional[list[str]] = None
    phenotype_groups: Optional[list[str]] = None
    gene_panels: Optional[list[str]] = None
    default_gene_panels: list[str] = []
    cohorts: Optional[list[str]] = None
    human_genome_build: str = None

    rank_model_version: Optional[str] = None
    rank_score_threshold: int = None
    sv_rank_model_version: Optional[str] = None
    analysis_date: Optional[datetime] = None
    samples: list[ScoutIndividual] = []

    delivery_report: Optional[str] = None
    coverage_qc_report: Optional[str] = None
    cnv_report: Optional[str] = None
    multiqc: Optional[str] = None
    track: Literal["rare", "cancer"] = "rare"

    model_config = ConfigDict(validate_assignment=True)


class BalsamicLoadConfig(ScoutLoadConfig):
    madeline: Optional[str] = None
    vcf_cancer: Annotated[str, BeforeValidator(field_not_none)] = None
    vcf_cancer_sv: Optional[str] = None
    vcf_cancer_research: Optional[str] = None
    vcf_cancer_sv_research: Optional[str] = None
    samples: list[ScoutCancerIndividual] = []


class BalsamicUmiLoadConfig(BalsamicLoadConfig):
    pass


class MipLoadConfig(ScoutLoadConfig):
    chromograph_image_files: Optional[list[str]] = None
    chromograph_prefixes: Optional[list[str]] = None
    madeline: Optional[str] = None
    peddy_check: Optional[str] = None
    peddy_ped: Optional[str] = None
    peddy_sex: Optional[str] = None
    samples: list[ScoutMipIndividual] = []
    smn_tsv: Optional[str] = None
    variant_catalog: Optional[str] = None
    vcf_mei: Optional[str] = None
    vcf_mei_research: Optional[str] = None
    vcf_snv: Annotated[str, BeforeValidator(field_not_none)] = None
    vcf_snv_research: Annotated[Optional[str], BeforeValidator(field_not_none)] = None
    vcf_str: Optional[str] = None
    vcf_sv: Annotated[Optional[str], BeforeValidator(field_not_none)] = None
    vcf_sv_research: Annotated[Optional[str], BeforeValidator(field_not_none)] = None


class RnafusionLoadConfig(ScoutLoadConfig):
    multiqc_rna: Optional[str] = None
    gene_fusion: Optional[str] = None
    gene_fusion_report_research: Optional[str] = None
    RNAfusion_inspector: Optional[str] = None
    RNAfusion_inspector_research: Optional[str] = None
    RNAfusion_report: Optional[str] = None
    RNAfusion_report_research: Optional[str] = None
    samples: list[ScoutCancerIndividual] = []



import logging
import tempfile
from datetime import datetime
from json.decoder import JSONDecodeError
from pathlib import Path
from typing import Optional

from cg.apps.crunchy.models import CrunchyFile, CrunchyMetadata
from cg.constants.constants import FileFormat
from cg.io.controller import ReadFile, ReadStream, WriteFile
from cg.utils.date import get_date

LOG = logging.getLogger(__name__)

NR_OF_FILES_IN_METADATA = 3


# Methods to get file information
def get_log_dir(file_path: Path) -> Path:
    """Return the path to where logs should be stored"""
    return file_path.parent


def get_fastq_to_spring_sbatch_path(log_dir: Path, run_name: str = None) -> Path:
    """Return the path to where compression sbatch should be printed"""
    return Path(log_dir, "_".join([run_name, "compress_fastq.sh"]))


def get_spring_to_fastq_sbatch_path(log_dir: Path, run_name: str = None) -> Path:
    """Return the path to where decompression sbatch should be printed"""
    return Path(log_dir, "_".join([run_name, "decompress_spring.sh"]))


def get_tmp_dir(prefix: str, suffix: str, base: str = None) -> str:
    """Create a temporary directory and return the path to it"""

    with tempfile.TemporaryDirectory(prefix=prefix, suffix=suffix, dir=base) as dir_name:
        tmp_dir_path = dir_name

    LOG.info("Created temporary dir %s", tmp_dir_path)
    return tmp_dir_path


def get_crunchy_metadata(metadata_path: Path) -> CrunchyMetadata:
    """Validate content of metadata file and return mapped content."""
    LOG.info(f"Fetch SPRING metadata from {metadata_path}")
    try:
        content: list[dict[str, str]] = ReadFile.get_content_from_file(
            file_format=FileFormat.JSON, file_path=metadata_path
        )
    except JSONDecodeError:
        message = "No content in SPRING metadata file"
        LOG.warning(message)
        raise SyntaxError(message)
    metadata = CrunchyMetadata(files=content)

    if len(metadata.files) != NR_OF_FILES_IN_METADATA:
        LOG.warning(f"Wrong number of files in SPRING metadata file: {metadata_path}")
        LOG.info(
            f"Found {len(metadata.files)} files, should always be {NR_OF_FILES_IN_METADATA} files"
        )
        raise SyntaxError

    return metadata


def get_file_updated_at(crunchy_metadata: CrunchyMetadata) -> Optional[datetime.date]:
    """Check if a SPRING metadata file has been updated and return the date when updated"""
    return crunchy_metadata.files[0].updated


def get_spring_archive_files(crunchy_metadata: CrunchyMetadata) -> dict[str, CrunchyFile]:
    """Map the files in SPRING metadata to a dictionary

    Returns: {
                "fastq_first" : {file_info},
                "fastq_second" : {file_info},
                "spring" : {file_info},
              }
    """
    names_map = {"first_read": "fastq_first", "second_read": "fastq_second", "spring": "spring"}
    archive_files = {}
    for file_info in crunchy_metadata.files:
        file_name = names_map[file_info.file]
        archive_files[file_name] = file_info
    return archive_files


def write_metadata_content(spring_metadata: CrunchyMetadata, spring_metadata_path: Path) -> None:
    """Update the file on disk."""
    content: dict = ReadStream.get_content_from_stream(
        file_format=FileFormat.JSON, stream=spring_metadata.json(exclude_none=True)
    )
    WriteFile.write_file_from_content(
        content=content["files"], file_format=FileFormat.JSON, file_path=spring_metadata_path
    )


def update_metadata_date(spring_metadata_path: Path) -> None:
    """Update date in the SPRING metadata file to today date"""
    now: datetime = get_date()
    spring_metadata: CrunchyMetadata = get_crunchy_metadata(spring_metadata_path)
    LOG.info("Adding today date to SPRING metadata file")
    for file_info in spring_metadata.files:
        file_info.updated = now.date()
    write_metadata_content(
        spring_metadata=spring_metadata, spring_metadata_path=spring_metadata_path
    )


def update_metadata_paths(spring_metadata_path: Path, new_parent_path: Path) -> None:
    """Update paths for file in SPRING metadata file."""
    spring_metadata: CrunchyMetadata = get_crunchy_metadata(metadata_path=spring_metadata_path)
    LOG.info(f"Updating file paths in SPRING metadata file: {spring_metadata_path}")
    for file in spring_metadata.files:
        file.path = Path(new_parent_path, Path(file.path).name)
    write_metadata_content(
        spring_metadata=spring_metadata, spring_metadata_path=spring_metadata_path
    )



from typing import Any, Optional, Union

from pydantic.v1 import ValidationError

from cg.apps.lims import LimsAPI
from cg.constants.invoice import CostCenters, CustomerNames
from cg.constants.priority import PriorityTerms
from cg.constants.sequencing import RecordType
from cg.models.invoice.invoice import (
    InvoiceApplication,
    InvoiceContact,
    InvoiceInfo,
    InvoiceReport,
)
from cg.server.ext import FlaskLims
from cg.server.ext import lims as genologics_lims
from cg.store import Store
from cg.store.models import Customer, Invoice, Pool, Sample, User


class InvoiceAPI:
    def __init__(self, db: Store, lims_api: LimsAPI, invoice_obj: Invoice):
        self.db = db
        self.lims_api = lims_api
        self.log = []
        self.invoice_obj = invoice_obj
        self.customer_obj = invoice_obj.customer
        self.record_type = ""
        self.raw_records = []
        self._set_record_type()
        self.genologics_lims: FlaskLims = genologics_lims
        self.invoice_info: Optional[InvoiceInfo] = None

    def _set_record_type(self):
        """Define the record_type based on the invoice object.
        It can only be either pool or sample."""
        if self.invoice_obj.pools:
            self.raw_records = self.invoice_obj.pools
            self.record_type = RecordType.Pool
        elif self.invoice_obj.samples:
            self.record_type = RecordType.Sample
            self.raw_records = self.invoice_obj.samples

    def get_customer_by_cost_center(self, cost_center: str) -> Union[Customer, str]:
        """Return the costumer based on cost center."""
        return (
            self.db.get_customer_by_internal_id(customer_internal_id=CustomerNames.cust999)
            if cost_center.lower() == CostCenters.kth
            else self.customer_obj
        )

    def get_customer_invoice_contact(self, customer: Customer, msg: str) -> Any:
        """Return the customer invoice contact."""
        if not customer.invoice_contact:
            self.log.append(msg)
            return None
        return customer.invoice_contact

    def get_contact(
        self, customer: Customer, customer_invoice_contact: Optional[User], msg: str
    ) -> Optional[InvoiceContact]:
        """Return the contact information."""
        try:
            contact = InvoiceContact(
                name=customer_invoice_contact.name,
                email=customer_invoice_contact.email,
                customer_name=customer.name,
                reference=customer.invoice_reference,
                address=customer.invoice_address,
            )
            return contact

        except ValidationError:
            self.log.append(msg)
            return None

    def get_contact_info(self, cost_center: str) -> Optional[InvoiceContact]:
        """Return contact info for a customer."""

        msg = (
            f"Could not open/generate invoice. Contact information missing in database for "
            f"customer {self.customer_obj.internal_id}. See log files."
        )
        customer = self.get_customer_by_cost_center(cost_center=cost_center)
        customer_invoice_contact = self.get_customer_invoice_contact(customer=customer, msg=msg)
        return self.get_contact(
            customer_invoice_contact=customer_invoice_contact, customer=customer, msg=msg
        )

    def get_invoice_report(self, cost_center: str) -> Optional[dict]:
        """Return invoice information as dictionary to generate Excel report."""

        records: list[dict] = []
        pooled_samples = []

        for raw_record in self.raw_records:
            if self.record_type == RecordType.Pool:
                pooled_samples += self.genologics_lims.samples_in_pools(
                    raw_record.name, raw_record.ticket
                )
            record = self.get_invoice_entity_record(
                cost_center=cost_center.lower(),
                discount=self.invoice_obj.discount,
                record=raw_record,
            )
            if record:
                records.append(record.dict())
            else:
                return None

        customer_obj = self.invoice_obj.customer
        contact = self.get_contact_info(cost_center)
        if not contact:
            return None

        try:
            invoice_report = InvoiceReport(
                cost_center=cost_center,
                project_number=getattr(customer_obj, f"project_account_{cost_center.lower()}"),
                customer_id=customer_obj.internal_id,
                customer_name=customer_obj.name,
                agreement=customer_obj.agreement_registration,
                invoice_id=self.invoice_obj.id,
                contact=contact.dict(),
                records=records,
                pooled_samples=pooled_samples,
                record_type=self.record_type,
            )
            return invoice_report.dict()
        except ValidationError:
            self.log.append("ValidationError in InvoiceReport class.")
            return None

    def _discount_price(self, record: Sample or Pool, discount: int = 0) -> Optional[int]:
        """Return discount price for a sample or pool."""
        priority = self.get_priority(record, for_discount_price=True)
        full_price = getattr(record.application_version, f"price_{priority}")
        discount_factor = 1 - discount / 100
        if not full_price:
            return None
        return round(full_price * discount_factor)

    def _cost_center_split_factor(
        self, price: int, cost_center: str, percent_kth: int, tag: str, version: str
    ) -> Optional[int]:
        """Return split price based on cost center."""
        if price:
            try:
                if cost_center == CostCenters.kth:
                    split_factor = percent_kth / 100
                else:
                    split_factor = 1 - percent_kth / 100
                split_price = round(price * split_factor)
            except ValueError:
                self.log.append(
                    f"Could not calculate price for samples with application "
                    f"tag/version: {tag}/{version}. Missing %KTH"
                )
                return None
        else:
            self.log.append(
                f"Could not get price for samples with application tag/version: {tag}/{version}."
            )
            return None
        return split_price

    def get_invoice_entity_record(
        self, cost_center: str, discount: int, record: Sample or Pool
    ) -> InvoiceInfo:
        """Return invoice information for a specific sample or pool."""
        application = self.get_application(record=record, discount=discount)

        split_discounted_price = self._cost_center_split_factor(
            price=application.discounted_price,
            cost_center=cost_center,
            percent_kth=application.percent_kth,
            tag=application.tag,
            version=application.version,
        )
        invoice_info = self.get_invoice_info(
            record=record,
            split_discounted_price=split_discounted_price,
            cost_center=cost_center,
            application=application,
        )

        return invoice_info

    def get_application(
        self, record: Sample or Pool, discount: int
    ) -> Optional[InvoiceApplication]:
        """Return the application information."""
        try:
            application = InvoiceApplication(
                tag=record.application_version.application.tag,
                version=record.application_version.version,
                percent_kth=record.application_version.application.percent_kth,
                discounted_price=self._discount_price(record, discount),
            )
            return application

        except ValidationError:
            self.log.append(f"Application tag/version seems to be missing for sample {record.id}.")

            return None

    def get_ticket(self, record: Sample or Pool) -> str:
        """Return ticket."""
        return record.ticket if self.record_type == RecordType.Pool else record.original_ticket

    def get_lims_id(self, record: Sample or Pool) -> str:
        """Return Lims id."""
        return None if self.record_type == RecordType.Pool else record.internal_id

    def get_priority(self, record: Sample or Pool, for_discount_price: bool = False) -> str:
        """Return the priority."""
        if self.customer_obj.internal_id == CustomerNames.cust032:
            priority = PriorityTerms.STANDARD
        elif self.record_type == RecordType.Pool or (
            record.priority_human == "clinical trials" and for_discount_price
        ):
            priority = PriorityTerms.RESEARCH
        else:
            priority = record.priority_human
        return priority

    def get_invoice_info(
        self, record, split_discounted_price: int, cost_center: str, application: InvoiceApplication
    ) -> InvoiceInfo:
        """Return the invoice_info to be used in the invoice report."""
        order = record.order
        ticket = self.get_ticket(record)
        lims_id = self.get_lims_id(record)
        priority = self.get_priority(record)
        price_kth: int = 0
        if cost_center == CostCenters.ki:
            price_kth: int = self._cost_center_split_factor(
                price=application.discounted_price,
                cost_center=CostCenters.kth,
                percent_kth=application.percent_kth,
                tag=application.tag,
                version=application.version,
            )

        try:
            invoice_info = InvoiceInfo(
                name=record.name,
                lims_id=lims_id,
                id=record.id,
                application_tag=record.application_version.application.tag,
                project=f"{order or 'NA'} ({ticket or 'NA'})",
                date=record.received_at.date() if record.received_at else "",
                price=split_discounted_price,
                priority=priority,
                price_kth=price_kth if cost_center == CostCenters.ki else None,
                total_price=application.discounted_price if cost_center == CostCenters.ki else None,
            )
            self.set_invoice_info(invoice_info=invoice_info)
            return invoice_info

        except ValidationError:
            self.log.append("ValidationError in InvoiceInfo class.")

    def set_invoice_info(self, invoice_info: InvoiceInfo):
        """Set invoice_info."""
        self.invoice_info = invoice_info

    def total_price(self) -> Optional[float]:
        """Return the total price for all records in the invoice."""

        discount = self.invoice_obj.discount
        total_price = 0

        for record in self.raw_records:
            discount_price = self._discount_price(record, discount)
            if discount_price:
                total_price += discount_price
            else:
                return None
        return total_price



"""Test for upload API."""
import datetime as dt

from cg.meta.upload.upload_api import UploadAPI
from cg.models.cg_config import CGConfig
from cg.store.models import Analysis, Family
from tests.cli.workflow.conftest import tb_api


def test_mip_dna_update_uploaded_at(mip_dna_context: CGConfig, mip_dna_case: Family):
    """Test setting uploaded at for a finished analysis."""
    # GIVEN an analysis that should be uploaded
    upload_api: UploadAPI = UploadAPI(
        config=mip_dna_context, analysis_api=mip_dna_context.meta_apis["analysis_api"]
    )
    dna_mip_analysis: Analysis = mip_dna_case.analyses[0]
    assert dna_mip_analysis.uploaded_at is None

    # WHEN setting the uploaded at
    upload_api.update_uploaded_at(analysis=dna_mip_analysis)

    # THEN the uploaded at should be set to a date
    assert isinstance(dna_mip_analysis.uploaded_at, dt.datetime)


def test_mip_rna_update_uploaded_at(mip_rna_context: CGConfig, mip_rna_analysis: Analysis):
    """Test setting uploaded at for a finished analysis."""
    # GIVEN an analysis that should be uploaded
    upload_api: UploadAPI = UploadAPI(
        config=mip_rna_context, analysis_api=mip_rna_context.meta_apis["analysis_api"]
    )

    assert mip_rna_analysis.uploaded_at is None

    # WHEN setting the uploaded at
    upload_api.update_uploaded_at(analysis=mip_rna_analysis)

    # THEN the uploaded at should be set to a date
    assert isinstance(mip_rna_analysis.uploaded_at, dt.datetime)



"""Chanjo API"""
import logging
import tempfile
from typing import Optional

from cg.constants.constants import FileFormat
from cg.io.controller import ReadStream
from cg.utils import Process

LOG = logging.getLogger(__name__)


class ChanjoAPI:

    """Interface to Chanjo, the coverage analysis tool"""

    def __init__(self, config: dict):
        self.chanjo_config = config["chanjo"]["config_path"]
        self.chanjo_binary = config["chanjo"]["binary_path"]
        self.process = Process(binary=self.chanjo_binary, config=self.chanjo_config)

    def upload(
        self, sample_id: str, sample_name: str, group_id: str, group_name: str, bed_file: str
    ):
        """Upload coverage for a sample"""

        load_parameters = [
            "load",
            "--sample",
            sample_id,
            "--name",
            sample_name,
            "--group",
            group_id,
            "--group-name",
            group_name,
            "--threshold",
            "10",
            bed_file,
        ]

        self.process.run_command(parameters=load_parameters)

    def sample(self, sample_id: str) -> Optional[dict]:
        """Fetch sample from the database"""

        sample_parameters = ["db", "samples", "-s", sample_id]
        self.process.run_command(parameters=sample_parameters)
        samples: list = ReadStream.get_content_from_stream(
            file_format=FileFormat.JSON, stream=self.process.stdout
        )
        for sample in samples:
            if sample["id"] == sample_id:
                return sample
        return None

    def delete_sample(self, sample_id: str):
        """Delete sample from database"""
        delete_parameters = ["db", "remove", sample_id]
        self.process.run_command(parameters=delete_parameters)

    def omim_coverage(self, samples: list[dict]) -> dict:
        """Calculate OMIM coverage for samples"""

        omim_parameters = ["calculate", "coverage", "--omim"]
        for sample in samples:
            omim_parameters.extend(["-s", sample["id"]])
        self.process.run_command(parameters=omim_parameters)
        return ReadStream.get_content_from_stream(
            file_format=FileFormat.JSON, stream=self.process.stdout
        )

    def sample_coverage(self, sample_id: str, panel_genes: list) -> dict:
        """Calculate coverage for samples"""

        with tempfile.NamedTemporaryFile(mode="w+t") as tmp_gene_file:
            tmp_gene_file.write("\n".join([str(gene) for gene in panel_genes]))
            tmp_gene_file.flush()
            coverage_parameters = [
                "calculate",
                "coverage",
                "-s",
                sample_id,
                "-f",
                tmp_gene_file.name,
            ]
            self.process.run_command(parameters=coverage_parameters)
        return ReadStream.get_content_from_stream(
            file_format=FileFormat.JSON, stream=self.process.stdout
        ).get(sample_id)



"""Module to create sbatch files and communicate with SLURM."""
import logging
from pathlib import Path
from typing import Optional

from cg.apps.slurm.sbatch import (
    DRAGEN_SBATCH_HEADER_TEMPLATE,
    SBATCH_BODY_TEMPLATE,
    SBATCH_HEADER_TEMPLATE,
)
from cg.constants.slurm import Slurm
from cg.models.slurm.sbatch import Sbatch
from cg.utils import Process

LOG = logging.getLogger(__name__)


class SlurmAPI:
    """API to SLURM."""

    def __init__(self):
        """Initialize SlurmAPI class."""
        self.process: Process = Process(binary="sbatch")
        self.dry_run: bool = False

    def set_dry_run(self, dry_run: bool) -> None:
        """Set dry run."""
        LOG.debug(f"SlurmAPI: Set dry run to {dry_run}")
        self.dry_run = dry_run

    @staticmethod
    def generate_sbatch_content(sbatch_parameters: Sbatch) -> str:
        """Take a parameters object and generate a string with sbatch information."""
        if hasattr(sbatch_parameters, Slurm.PARTITION.value):
            sbatch_header: str = SlurmAPI.generate_dragen_sbatch_header(
                sbatch_parameters=sbatch_parameters
            )
        else:
            sbatch_header: str = SlurmAPI.generate_sbatch_header(
                sbatch_parameters=sbatch_parameters
            )
        sbatch_body: str = SlurmAPI.generate_sbatch_body(
            commands=sbatch_parameters.commands, error_function=sbatch_parameters.error
        )
        return "\n".join([sbatch_header, sbatch_body])

    @staticmethod
    def generate_sbatch_header(sbatch_parameters: Sbatch) -> str:
        return SBATCH_HEADER_TEMPLATE.format(**sbatch_parameters.model_dump())

    @staticmethod
    def generate_dragen_sbatch_header(sbatch_parameters: Sbatch) -> str:
        return DRAGEN_SBATCH_HEADER_TEMPLATE.format(**sbatch_parameters.model_dump())

    @staticmethod
    def generate_sbatch_body(commands: str, error_function: Optional[str] = None) -> str:
        if not error_function:
            error_function = "log 'Something went wrong, aborting'"

        return SBATCH_BODY_TEMPLATE.format(**{"error_body": error_function, "commands": commands})

    @staticmethod
    def write_sbatch_file(sbatch_content: str, sbatch_path: Path, dry_run: bool) -> None:
        if dry_run:
            LOG.info("Write sbatch content to path %s: \n%s", sbatch_path, sbatch_content)
            return
        LOG.debug("Write sbatch content %s to %s", sbatch_content, sbatch_path)
        with open(sbatch_path, mode="w+t") as sbatch_file:
            sbatch_file.write(sbatch_content)

    def submit_sbatch_job(self, sbatch_path: Path) -> int:
        LOG.info("Submit sbatch %s", sbatch_path)
        sbatch_parameters: list[str] = [str(sbatch_path)]
        self.process.run_command(parameters=sbatch_parameters, dry_run=self.dry_run)
        if self.process.stderr:
            LOG.info(self.process.stderr)
        LOG.info(self.process.stdout)
        try:
            job_number: int = int(self.process.stdout.strip().split()[-1])
        except (ValueError, IndexError):
            LOG.warning("Could not get slurm job number")
            job_number = 0
            if self.dry_run:
                job_number = 123456
        return job_number

    def submit_sbatch(self, sbatch_content: str, sbatch_path: Path) -> int:
        """Submit sbatch file to slurm.

        Return the slurm job id
        """
        SlurmAPI.write_sbatch_file(
            sbatch_content=sbatch_content, sbatch_path=sbatch_path, dry_run=self.dry_run
        )
        return self.submit_sbatch_job(sbatch_path=sbatch_path)



"""Module for Rnafusion Analysis API."""

import logging
from pathlib import Path
from typing import Any, Optional

from cg import resources
from cg.constants import Pipeline
from cg.constants.constants import FileFormat, Strandedness
from cg.constants.rnafusion import RNAFUSION_METRIC_CONDITIONS
from cg.constants.tb import AnalysisStatus
from cg.exc import CgError, MetricsQCError, MissingMetrics
from cg.io.controller import ReadFile, WriteFile
from cg.io.json import read_json
from cg.meta.workflow.nf_analysis import NfAnalysisAPI
from cg.models.cg_config import CGConfig
from cg.models.deliverables.metric_deliverables import (
    MetricsBase,
    MetricsDeliverablesCondition,
    MultiqcDataJson,
)
from cg.models.nf_analysis import PipelineDeliverables
from cg.models.rnafusion.rnafusion import (
    RnafusionAnalysis,
    RnafusionParameters,
    RnafusionSampleSheetEntry,
)
from cg.store.models import Family, Sample

LOG = logging.getLogger(__name__)


class RnafusionAnalysisAPI(NfAnalysisAPI):
    """Handles communication between RNAFUSION processes
    and the rest of CG infrastructure."""

    def __init__(
        self,
        config: CGConfig,
        pipeline: Pipeline = Pipeline.RNAFUSION,
    ):
        super().__init__(config=config, pipeline=pipeline)
        self.root_dir: str = config.rnafusion.root
        self.nfcore_pipeline_path: str = config.rnafusion.pipeline_path
        self.references: str = config.rnafusion.references
        self.profile: str = config.rnafusion.profile
        self.conda_env: str = config.rnafusion.conda_env
        self.conda_binary: str = config.rnafusion.conda_binary
        self.tower_binary_path: str = config.rnafusion.tower_binary_path
        self.tower_pipeline: str = config.rnafusion.tower_pipeline
        self.account: str = config.rnafusion.slurm.account
        self.email: str = config.rnafusion.slurm.mail_user
        self.compute_env: str = config.rnafusion.compute_env
        self.revision: str = config.rnafusion.revision
        self.nextflow_binary_path: str = config.rnafusion.binary_path

    @property
    def use_read_count_threshold(self) -> bool:
        """Defines whether the threshold for adequate read count should be passed for all samples
        when determining if the analysis for a case should be automatically started."""
        return True

    @staticmethod
    def get_deliverables_template_content() -> list[dict]:
        """Return deliverables file template content."""
        return ReadFile.get_content_from_file(
            file_format=FileFormat.YAML,
            file_path=resources.RNAFUSION_BUNDLE_FILENAMES_PATH,
        )

    def get_sample_sheet_content_per_sample(
        self, sample: Sample, case_id: str, strandedness: Strandedness
    ) -> list[list[str]]:
        """Get sample sheet content per sample."""
        sample_metadata: list[dict] = self.gather_file_metadata_for_sample(sample_obj=sample)
        fastq_forward_read_paths: list[str] = self.extract_read_files(
            metadata=sample_metadata, forward_read=True
        )
        fastq_reverse_read_paths: list[str] = self.extract_read_files(
            metadata=sample_metadata, reverse_read=True
        )

        sample_sheet_entry = RnafusionSampleSheetEntry(
            name=case_id,
            fastq_forward_read_paths=fastq_forward_read_paths,
            fastq_reverse_read_paths=fastq_reverse_read_paths,
            strandedness=strandedness,
        )
        return sample_sheet_entry.reformat_sample_content()

    def get_sample_sheet_content(self, case_id: str, strandedness: Strandedness) -> list[list[Any]]:
        """Returns content for sample sheet."""
        case: Family = self.status_db.get_case_by_internal_id(internal_id=case_id)
        if len(case.links) != 1:
            raise NotImplementedError(
                "Case objects are assumed to be related to a single sample (one link)"
            )
        LOG.debug("Getting sample sheet information")
        for link in case.links:
            content_per_sample = self.get_sample_sheet_content_per_sample(
                sample=link.sample, case_id=case_id, strandedness=strandedness
            )
            return content_per_sample

    def get_pipeline_parameters(
        self, case_id: str, genomes_base: Optional[Path] = None
    ) -> RnafusionParameters:
        """Get Rnafusion parameters."""
        LOG.debug("Getting parameters information")
        return RnafusionParameters(
            cluster_options=f"--qos={self.get_slurm_qos_for_case(case_id=case_id)}",
            genomes_base=genomes_base or self.get_references_path(),
            sample_sheet_path=self.get_sample_sheet_path(case_id=case_id),
            outdir=self.get_case_path(case_id=case_id),
            priority=self.account,
        )

    def get_references_path(self, genomes_base: Optional[Path] = None) -> Path:
        if genomes_base:
            return genomes_base.absolute()
        return Path(self.references).absolute()

    def config_case(
        self,
        case_id: str,
        strandedness: Strandedness,
        genomes_base: Path,
        dry_run: bool,
    ) -> None:
        """Create config files (parameters and sample sheet) for Rnafusion analysis."""
        self.create_case_directory(case_id=case_id, dry_run=dry_run)
        sample_sheet_content: list[list[Any]] = self.get_sample_sheet_content(
            case_id=case_id, strandedness=strandedness
        )
        pipeline_parameters: RnafusionParameters = self.get_pipeline_parameters(
            case_id=case_id, genomes_base=genomes_base
        )
        if dry_run:
            LOG.info("Dry run: Config files will not be written")
            return
        self.write_sample_sheet(
            content=sample_sheet_content,
            file_path=self.get_sample_sheet_path(case_id=case_id),
            header=RnafusionSampleSheetEntry.headers(),
        )
        self.write_params_file(case_id=case_id, pipeline_parameters=pipeline_parameters.dict())

    def report_deliver(self, case_id: str) -> None:
        """Create deliverables file."""
        deliverables_content: PipelineDeliverables = self.get_deliverables_for_case(case_id=case_id)
        self.write_deliverables_file(
            deliverables_content=deliverables_content.dict(),
            file_path=self.get_deliverables_file_path(case_id=case_id),
        )
        LOG.info(
            f"Writing deliverables file in {self.get_deliverables_file_path(case_id=case_id).as_posix()}"
        )

    def get_multiqc_json_path(self, case_id: str) -> Path:
        """Return the path of the multiqc_data.json file."""
        return Path(self.root_dir, case_id, "multiqc", "multiqc_data", "multiqc_data.json")

    def get_multiqc_json_metrics(self, case_id: str) -> list[MetricsBase]:
        """Get a multiqc_data.json file and returns metrics and values formatted."""
        case: Family = self.status_db.get_case_by_internal_id(internal_id=case_id)
        sample_id: str = case.links[0].sample.internal_id
        multiqc_json: MultiqcDataJson = MultiqcDataJson(
            **read_json(file_path=self.get_multiqc_json_path(case_id=case_id))
        )
        metrics_values: dict = {}
        for key in multiqc_json.report_general_stats_data:
            if case_id in key:
                metrics_values.update(list(key.values())[0])
        return [
            MetricsBase(
                header=None,
                id=sample_id,
                input="multiqc_data.json",
                name=metric_name,
                step="multiqc",
                value=metric_value,
                condition=RNAFUSION_METRIC_CONDITIONS.get(metric_name, None),
            )
            for metric_name, metric_value in metrics_values.items()
        ]

    @staticmethod
    def ensure_mandatory_metrics_present(metrics: list[MetricsBase]) -> None:
        """Check that all mandatory metrics are present. Raise error if missing."""
        given_metrics: set = {metric.name for metric in metrics}
        mandatory_metrics: set = set(RNAFUSION_METRIC_CONDITIONS.keys())
        missing_metrics: set = mandatory_metrics.difference(given_metrics)
        if missing_metrics:
            LOG.error(f"Some mandatory metrics are missing: {', '.join(missing_metrics)}")
            raise MissingMetrics()

    def write_metrics_deliverables(self, case_id: str, dry_run: bool = False) -> None:
        """Write <case>_metrics_deliverables.yaml file."""
        metrics_deliverables_path: Path = self.get_metrics_deliverables_path(case_id=case_id)
        metrics = self.get_multiqc_json_metrics(case_id=case_id)
        self.ensure_mandatory_metrics_present(metrics=metrics)

        if dry_run:
            LOG.info(
                f"Dry-run: metrics deliverables file would be written to {metrics_deliverables_path.as_posix()}"
            )
            return

        LOG.info(f"Writing metrics deliverables file to {metrics_deliverables_path.as_posix()}")
        WriteFile.write_file_from_content(
            content={"metrics": [metric.dict() for metric in metrics]},
            file_format=FileFormat.YAML,
            file_path=metrics_deliverables_path,
        )

    def validate_qc_metrics(self, case_id: str, dry_run: bool = False) -> None:
        """Validate the information from a qc metrics deliverable file."""

        if dry_run:
            LOG.info("Dry-run: QC metrics validation would be performed")
            return

        LOG.info("Validating QC metrics")
        try:
            metrics_deliverables_path: Path = self.get_metrics_deliverables_path(case_id=case_id)
            qc_metrics_raw: dict = ReadFile.get_content_from_file(
                file_format=FileFormat.YAML, file_path=metrics_deliverables_path
            )
            MetricsDeliverablesCondition(**qc_metrics_raw)
        except MetricsQCError as error:
            LOG.error(f"QC metrics failed for {case_id}")
            self.trailblazer_api.set_analysis_status(case_id=case_id, status=AnalysisStatus.FAILED)
            self.trailblazer_api.add_comment(case_id=case_id, comment=str(error))
            raise MetricsQCError from error
        except CgError as error:
            LOG.error(f"Could not create metrics deliverables file: {error}")
            self.trailblazer_api.set_analysis_status(case_id=case_id, status=AnalysisStatus.ERROR)
            raise CgError from error
        self.trailblazer_api.set_analysis_status(case_id=case_id, status=AnalysisStatus.COMPLETED)

    def parse_analysis(self, qc_metrics_raw: list[MetricsBase], **kwargs) -> RnafusionAnalysis:
        """Parse Rnafusion output analysis files and return analysis model."""
        sample_metrics: dict[str, dict] = {}
        for metric in qc_metrics_raw:
            metric.name = metric.name.replace("5_3_bias", "bias_5_3")
            try:
                sample_metrics[metric.id].update({metric.name.lower(): metric.value})
            except KeyError:
                sample_metrics[metric.id] = {metric.name.lower(): metric.value}
        return RnafusionAnalysis(sample_metrics=sample_metrics)

    def get_latest_metadata(self, case_id: str) -> RnafusionAnalysis:
        """Return the latest metadata of a specific Rnafusion case."""
        qc_metrics: list[MetricsBase] = self.get_multiqc_json_metrics(case_id)
        return self.parse_analysis(qc_metrics_raw=qc_metrics)



"""API for encryption on Hasta"""
import logging
import subprocess
from io import TextIOWrapper
from pathlib import Path
from tempfile import NamedTemporaryFile
from typing import Optional, Union

from cg.apps.slurm.slurm_api import SlurmAPI
from cg.constants import SPACE, FileExtensions
from cg.constants.encryption import EncryptionUserID, GPGParameters
from cg.constants.priority import SlurmQos
from cg.exc import ChecksumFailedError, FlowCellEncryptionError, FlowCellError
from cg.meta.encryption.sbatch import (
    FLOW_CELL_ENCRYPT_COMMANDS,
    FLOW_CELL_ENCRYPT_ERROR,
)
from cg.meta.tar.tar import TarAPI
from cg.models.flow_cell.flow_cell import FlowCellDirectoryData
from cg.models.slurm.sbatch import Sbatch
from cg.utils import Process
from cg.utils.checksum.checksum import sha512_checksum

LOG = logging.getLogger(__name__)
LIMIT_PIGZ_TASK: int = 3


class EncryptionAPI:
    """Class that uses gpg for various encryption and decryption functionality"""

    def __init__(self, binary_path: str, dry_run: bool = False):
        self.binary_path: str = binary_path
        self.process: Process = Process(binary=binary_path)
        self.dry_run: bool = dry_run

    def run_gpg_command(self, command: list) -> None:
        """Runs a GPG command"""
        LOG.info("Starting GPG command:")
        LOG.info(f"{self.process.binary} {' '.join(command)}")
        self.process.run_command(command, dry_run=self.dry_run)

    def run_passhprase_process(
        self, quality_level: int, count: int, passphrase_file: TextIOWrapper
    ) -> None:
        with open(passphrase_file.name, "w") as f:
            process = subprocess.Popen(
                [
                    self.binary_path,
                    "--gen-random",
                    str(quality_level),
                    str(count),
                ],
                stdout=f,
            )
        process.wait()

    def generate_temporary_passphrase_file(self, quality_level: int = 2, count: int = 256) -> Path:
        """Generates a temporary file with random bytes. The default values are based on the
        encryption parameters as used on the NAS:es connected to the sequencers"""
        LOG.debug("Generate temporary passphrase file:")
        passphrase_file = NamedTemporaryFile(delete=False)
        self.run_passhprase_process(
            passphrase_file=passphrase_file, quality_level=quality_level, count=count
        )
        LOG.debug(f"Temporary passphrase file generated: {passphrase_file.name}")

        return Path(passphrase_file.name)

    def get_symmetric_passphrase_cmd(
        self, passphrase_file_path: Path, quality_level: int = 2, count: int = 256
    ) -> str:
        """Return command to generate a symmetrical passphrase file."""
        return " ".join(
            [
                self.binary_path,
                "--gen-random",
                str(quality_level),
                str(count),
                ">",
                passphrase_file_path.as_posix(),
            ]
        )

    def get_asymmetrically_encrypt_passphrase_cmd(self, passphrase_file_path: Path) -> str:
        """Return command to asymmetrically encrypt a symmetrical passphrase file."""
        return " ".join(
            [
                self.binary_path,
                "--encrypt",
                "--recipient",
                EncryptionUserID.HASTA_USER_ID,
                "--output",
                passphrase_file_path.with_suffix(FileExtensions.GPG).as_posix(),
                passphrase_file_path.as_posix(),
            ]
        )

    def get_asymmetric_encryption_command(self, input_file: Path, output_file: Path) -> list[str]:
        """Generates the gpg command for asymmetric encryption"""
        encryption_parameters: list = GPGParameters.ASYMMETRIC_ENCRYPTION.copy()
        output_parameter: list = GPGParameters.OUTPUT_PARAMETER.copy()
        output_parameter.extend([str(output_file), str(input_file)])
        encryption_parameters.extend(output_parameter)
        return encryption_parameters

    def get_symmetric_encryption_command(self, input_file: Path, output_file: Path) -> list[str]:
        """Generates the gpg command for symmetric encryption of spring files"""
        encryption_parameters: list = GPGParameters.SYMMETRIC_ENCRYPTION.copy()
        encryption_parameters.append(str(self.temporary_passphrase))
        output_parameter: list = GPGParameters.OUTPUT_PARAMETER.copy()
        output_parameter.extend([str(output_file), str(input_file)])
        encryption_parameters.extend(output_parameter)
        return encryption_parameters

    def get_asymmetric_decryption_command(self, input_file: Path, output_file: Path) -> list[str]:
        """Generates the gpg command for asymmetric decryption"""
        decryption_parameters: list = GPGParameters.ASYMMETRIC_DECRYPTION.copy()
        output_parameter: list = GPGParameters.OUTPUT_PARAMETER.copy()
        output_parameter.extend([str(output_file), str(input_file)])
        decryption_parameters.extend(output_parameter)
        return decryption_parameters

    def get_symmetric_decryption_command(
        self, input_file: Path, output_file: Path, encryption_key: Path
    ) -> list[str]:
        """Generates the gpg command for symmetric decryption"""
        decryption_parameters: list = GPGParameters.SYMMETRIC_DECRYPTION.copy()
        decryption_parameters.append(str(encryption_key))
        output_parameter: list = GPGParameters.OUTPUT_PARAMETER.copy()
        output_parameter.extend([str(output_file), str(input_file)])
        decryption_parameters.extend(output_parameter)
        return decryption_parameters

    def create_pending_file(self, pending_path: Path) -> None:
        """Create a pending flag file."""
        LOG.info(f"Creating pending flag {pending_path}")
        if self.dry_run:
            return
        pending_path.touch(exist_ok=False)


class FlowCellEncryptionAPI(EncryptionAPI):
    """Encryption functionality for flow cells."""

    def __init__(
        self,
        binary_path: str,
        encryption_dir: Path,
        flow_cell: FlowCellDirectoryData,
        pigz_binary_path: str,
        sbatch_parameter: dict[str, Union[str, int]],
        slurm_api: SlurmAPI,
        tar_api: TarAPI,
        dry_run: bool = False,
    ):
        super().__init__(binary_path=binary_path, dry_run=dry_run)
        self.flow_cell: FlowCellDirectoryData = flow_cell
        self.encryption_dir: Path = encryption_dir
        self.tar_api: TarAPI = tar_api
        self.pigz_binary_path: str = pigz_binary_path
        self.slurm_api: SlurmAPI = slurm_api
        self.slurm_account: str = sbatch_parameter.get("account")
        self.slurm_hours: int = sbatch_parameter.get("hours")
        self.slurm_mail_user: str = sbatch_parameter.get("mail_user")
        self.slurm_memory: int = sbatch_parameter.get("memory")
        self.slurm_number_tasks: int = sbatch_parameter.get("number_tasks")
        self.slurm_api.set_dry_run(dry_run=self.dry_run)

    @property
    def flow_cell_encryption_dir(self) -> Path:
        return Path(self.encryption_dir, self.flow_cell.full_name)

    @property
    def flow_cell_encrypt_file_path_prefix(self) -> Path:
        return Path(self.flow_cell_encryption_dir, self.flow_cell.id)

    @property
    def pending_file_path(self) -> Path:
        return Path(self.flow_cell_encrypt_file_path_prefix.with_suffix(FileExtensions.PENDING))

    @property
    def complete_file_path(self) -> Path:
        return Path(self.flow_cell_encrypt_file_path_prefix.with_suffix(FileExtensions.COMPLETE))

    @property
    def encrypted_gpg_file_path(self) -> Path:
        return Path(
            self.flow_cell_encrypt_file_path_prefix.with_suffix(
                f"{FileExtensions.TAR}{FileExtensions.GZIP}{FileExtensions.GPG}"
            )
        )

    @property
    def encrypted_md5sum_file_path(self) -> Path:
        return Path(
            self.flow_cell_encrypt_file_path_prefix.with_suffix(
                f"{FileExtensions.TAR}{FileExtensions.GZIP}{FileExtensions.MD5SUM}"
            )
        )

    @property
    def decrypted_md5sum_file_path(self) -> Path:
        return Path(
            self.flow_cell_encrypt_file_path_prefix.with_suffix(
                f"{FileExtensions.TAR}{FileExtensions.GZIP}.degpg{FileExtensions.MD5SUM}"
            )
        )

    @property
    def symmetric_passphrase_file_path(self) -> Path:
        return Path(self.flow_cell_encrypt_file_path_prefix.with_suffix(FileExtensions.PASS_PHRASE))

    @property
    def final_passphrase_file_path(self) -> Path:
        return Path(
            self.flow_cell_encrypt_file_path_prefix.with_suffix(f".key{FileExtensions.GPG}")
        )

    def get_flow_cell_symmetric_encryption_command(
        self, output_file: Path, passphrase_file_path: Path
    ) -> str:
        """Generates the Gpg command for symmetric encryption of file."""
        encryption_parameters: list[str] = (
            [self.binary_path]
            + GPGParameters.SYMMETRIC_ENCRYPTION
            + [passphrase_file_path.as_posix()]
            + GPGParameters.OUTPUT_PARAMETER
            + [output_file.as_posix()]
        )
        return SPACE.join(encryption_parameters)

    def get_flow_cell_symmetric_decryption_command(
        self, input_file: Path, passphrase_file_path: Path
    ) -> str:
        """Generates the Gpg command for symmetric decryption."""
        decryption_parameters: list[str] = (
            [self.binary_path]
            + GPGParameters.SYMMETRIC_DECRYPTION
            + [passphrase_file_path.as_posix(), input_file.as_posix()]
        )
        return SPACE.join(decryption_parameters)

    def is_encryption_possible(self) -> Optional[bool]:
        """Check if requirements for encryption are meet.
        Raises:
            FlowCellError if sequencing is not ready, encryption is pending or complete.
        """
        if not self.flow_cell.is_flow_cell_ready():
            raise FlowCellError(f"Flow cell: {self.flow_cell.id} is not ready")
        if self.complete_file_path.exists():
            raise FlowCellEncryptionError(
                f"Encryption already completed for flow cell: {self.flow_cell.id}"
            )
        if self.pending_file_path.exists():
            raise FlowCellEncryptionError(
                f"Encryption already started for flow cell: {self.flow_cell.id}"
            )
        return True

    def encrypt_flow_cell(
        self,
    ) -> None:
        """Encrypt flow cell via GPG and SLURM."""
        error_function: str = FLOW_CELL_ENCRYPT_ERROR.format(
            pending_file_path=self.pending_file_path
        )
        commands: str = FLOW_CELL_ENCRYPT_COMMANDS.format(
            symmetric_passphrase=self.get_symmetric_passphrase_cmd(
                passphrase_file_path=self.symmetric_passphrase_file_path
            ),
            asymmetrically_encrypt_passphrase=self.get_asymmetrically_encrypt_passphrase_cmd(
                passphrase_file_path=self.symmetric_passphrase_file_path
            ),
            tar_encrypt_flow_cell_dir=self.tar_api.get_compress_cmd(input_path=self.flow_cell.path),
            parallel_gzip=f"{self.pigz_binary_path} -p {self.slurm_number_tasks - LIMIT_PIGZ_TASK} --fast -c",
            tee=f"tee >(md5sum > {self.encrypted_md5sum_file_path})",
            flow_cell_symmetric_encryption=self.get_flow_cell_symmetric_encryption_command(
                output_file=self.encrypted_gpg_file_path,
                passphrase_file_path=self.symmetric_passphrase_file_path,
            ),
            flow_cell_symmetric_decryption=self.get_flow_cell_symmetric_decryption_command(
                input_file=self.encrypted_gpg_file_path,
                passphrase_file_path=self.symmetric_passphrase_file_path,
            ),
            md5sum=f"md5sum > {self.decrypted_md5sum_file_path}",
            diff=f"diff -q {self.encrypted_md5sum_file_path} {self.decrypted_md5sum_file_path}",
            mv_passphrase_file=f"mv {self.symmetric_passphrase_file_path.with_suffix(FileExtensions.GPG)} {self.final_passphrase_file_path}",
            remove_pending_file=f"rm -f {self.pending_file_path}",
            flag_as_complete=f"touch {self.complete_file_path}",
        )
        sbatch_parameters = Sbatch(
            account=self.slurm_account,
            commands=commands,
            email=self.slurm_mail_user,
            error=error_function,
            hours=self.slurm_hours,
            job_name="_".join([self.flow_cell.id, "flow_cell_encryption"]),
            log_dir=self.flow_cell_encryption_dir.as_posix(),
            memory=self.slurm_memory,
            number_tasks=self.slurm_number_tasks,
            quality_of_service=SlurmQos.HIGH,
        )
        sbatch_content: str = self.slurm_api.generate_sbatch_content(sbatch_parameters)
        sbatch_path = Path(
            self.flow_cell_encrypt_file_path_prefix.with_suffix(FileExtensions.SBATCH)
        )
        sbatch_number: int = self.slurm_api.submit_sbatch(
            sbatch_content=sbatch_content, sbatch_path=sbatch_path
        )
        LOG.info(f"Flow cell encryption running as job {sbatch_number}")

    def start_encryption(self) -> None:
        """Check if requirements for starting encryption are meet if so starts encryption."""
        self.is_encryption_possible()
        self.flow_cell_encryption_dir.mkdir(exist_ok=True, parents=True)
        self.create_pending_file(pending_path=self.pending_file_path)
        self.encrypt_flow_cell()


class SpringEncryptionAPI(EncryptionAPI):
    """Encryption functionality for spring files"""

    def __init__(
        self,
        binary_path: str,
        dry_run: bool = False,
    ):
        super().__init__(binary_path=binary_path, dry_run=dry_run)
        self._temporary_passphrase = None

    def spring_symmetric_encryption(self, spring_file_path: Path) -> None:
        """Symmetrically encrypts a spring file"""
        output_file = self.encrypted_spring_file_path(spring_file_path)
        LOG.debug("*** ENCRYPTING SPRING FILE ***")
        LOG.info("Encrypt spring file: %s", spring_file_path)
        LOG.info("to output file     : %s", output_file)
        encryption_command: list = self.get_symmetric_encryption_command(
            input_file=spring_file_path, output_file=output_file
        )
        self.run_gpg_command(encryption_command)

    def key_asymmetric_encryption(self, spring_file_path: Path) -> None:
        """Asymmetrically encrypts the key used for spring file encryption"""
        output_file = self.encrypted_key_path(spring_file_path)
        LOG.debug("*** ENCRYPTING KEY FILE ***")
        LOG.info("Encrypt key file: %s", self.temporary_passphrase)
        LOG.info("to target file  : %s", output_file)
        encryption_command: list = self.get_asymmetric_encryption_command(
            input_file=self.temporary_passphrase, output_file=output_file
        )
        self.run_gpg_command(encryption_command)

    def spring_symmetric_decryption(self, spring_file_path: Path, output_file: Path) -> None:
        """decrypt a spring file"""
        input_file = self.encrypted_spring_file_path(spring_file_path)
        LOG.debug("*** DECRYPTING SPRING FILE ***")
        LOG.info("Decrypt spring file: %s", input_file)
        LOG.info("to target file     : %s", output_file)
        decryption_command: list = self.get_symmetric_decryption_command(
            input_file=input_file,
            output_file=output_file,
            encryption_key=self.encryption_key(spring_file_path),
        )
        self.run_gpg_command(decryption_command)

    def key_asymmetric_decryption(self, spring_file_path: Path) -> None:
        """Asymmetrically decrypts the key used for spring file decryption"""
        input_file = self.encrypted_key_path(spring_file_path)
        output_file = self.encryption_key(spring_file_path)
        LOG.debug("*** DECRYPTING KEY FILE ***")
        LOG.info("Decrypt key file: %s", input_file)
        LOG.info("to target file  : %s", output_file)
        decryption_command: list = self.get_asymmetric_decryption_command(
            input_file=input_file, output_file=output_file
        )
        self.run_gpg_command(decryption_command)

    def cleanup(self, spring_file_path: Path) -> None:
        """Removes any existing encrypted key file, key file and encrypted spring file before
        attempting encryption or after an error has occurred"""
        LOG.debug("*** PRE OR POST ENCRYPTION CLEANUP ***")
        try:
            self.decrypted_spring_file_checksum(spring_file_path).unlink()
            LOG.info("Removed existing decrypted checksum spring file")
        except FileNotFoundError:
            LOG.info("No decrypted checksum spring file to clean up, continuing cleanup")
        try:
            self.encrypted_spring_file_path(spring_file_path).unlink()
            LOG.info("Removed existing encrypted spring file")
        except FileNotFoundError:
            LOG.info("No encrypted spring file to clean up, continuing cleanup")
        try:
            self.encrypted_key_path(spring_file_path).unlink()
            LOG.info("Removed existing encrypted key file")
        except FileNotFoundError:
            LOG.info("No encrypted key file to clean up, continuing cleanup")
        try:
            self.encryption_key(spring_file_path).unlink()
            LOG.info("Removed existing key file")
        except FileNotFoundError:
            LOG.info("No existing key file to clean up, cleanup process completed")

    def compare_spring_file_checksums(self, spring_file_path: Path) -> None:
        """Make sure the encrypted spring file is correct and can be decrypted"""
        LOG.debug("PERFORMING CHECKSUM COMPARISON")
        if self.dry_run:
            LOG.info("Dry run, skipping checksum!")
            return
        self.key_asymmetric_decryption(spring_file_path=spring_file_path)
        self.spring_symmetric_decryption(
            spring_file_path=spring_file_path,
            output_file=self.decrypted_spring_file_checksum(spring_file_path),
        )
        is_checksum_equal = sha512_checksum(spring_file_path) == sha512_checksum(
            self.decrypted_spring_file_checksum(spring_file_path)
        )
        if not is_checksum_equal:
            raise ChecksumFailedError("Checksum comparison failed!")
        LOG.info("Checksum comparison successful!")

    def encrypted_key_path(self, spring_file_path: Path) -> Path:
        """The name of the encrypted key file"""
        return Path(spring_file_path).with_suffix(FileExtensions.KEY + FileExtensions.GPG)

    def encryption_key(self, spring_file_path: Path) -> Path:
        """The name of the encryption key"""
        return Path(spring_file_path).with_suffix(FileExtensions.KEY)

    def encrypted_spring_file_path(self, spring_file_path: Path) -> Path:
        """The name of the encrypted spring file"""
        return spring_file_path.with_suffix(FileExtensions.SPRING + FileExtensions.GPG)

    def decrypted_spring_file_checksum(self, spring_file_path: Path) -> Path:
        """The name of the decrypted spring file used to perform the checksum check"""
        return spring_file_path.with_suffix(FileExtensions.SPRING + FileExtensions.TMP)

    @property
    def temporary_passphrase(self) -> Path:
        """The name of the temporary passphrase"""
        if self._temporary_passphrase is None:
            self._temporary_passphrase: Path = self.generate_temporary_passphrase_file()
        return self._temporary_passphrase



from typing import Optional

import coloredlogs
import requests
from flask import Flask, redirect, session, url_for
from flask_admin.base import AdminIndexView
from flask_dance.consumer import oauth_authorized
from flask_dance.contrib.google import google, make_google_blueprint
from sqlalchemy.orm import scoped_session

from cg.store.database import get_scoped_session_registry
from cg.store.models import (
    Analysis,
    Application,
    ApplicationLimitations,
    ApplicationVersion,
    Bed,
    BedVersion,
    Collaboration,
    Customer,
    Delivery,
    Family,
    FamilySample,
    Flowcell,
    Invoice,
    Organism,
    Panel,
    Pool,
    Sample,
    SampleLaneSequencingMetrics,
    User,
)

from . import admin, api, ext, invoices


def create_app():
    """Generate a flask application."""
    app = Flask(__name__, template_folder="templates")
    _load_config(app)
    _configure_extensions(app)
    _register_blueprints(app)
    _register_teardowns(app)

    return app


def _load_config(app: Flask):
    app.config.from_object(__name__.replace("app", "config"))


def _configure_extensions(app: Flask):
    _initialize_logging(app)
    certs_resp = requests.get("https://www.googleapis.com/oauth2/v1/certs")
    app.config["GOOGLE_OAUTH_CERTS"] = certs_resp.json()

    ext.cors.init_app(app)
    ext.csrf.init_app(app)
    ext.db.init_app(app)
    ext.lims.init_app(app)
    if app.config["OSTICKET_API_KEY"]:
        ext.osticket.init_app(app)
    ext.admin.init_app(app, index_view=AdminIndexView(endpoint="admin"))
    app.json_provider_class = ext.CustomJSONEncoder


def _initialize_logging(app):
    coloredlogs.install(level="DEBUG" if app.debug else "INFO")


def _register_blueprints(app: Flask):
    if not app.config["CG_ENABLE_ADMIN"]:
        return

    oauth_bp = make_google_blueprint(
        client_id=app.config["GOOGLE_OAUTH_CLIENT_ID"],
        client_secret=app.config["GOOGLE_OAUTH_CLIENT_SECRET"],
        scope=["openid", "https://www.googleapis.com/auth/userinfo.email"],
    )

    @oauth_authorized.connect_via(oauth_bp)
    def logged_in(blueprint, token):
        """Called when the user logs in via Google OAuth."""
        resp = google.get("/oauth2/v1/userinfo?alt=json")
        assert resp.ok, resp.text
        user_data = resp.json()
        session["user_email"] = user_data["email"]

    app.register_blueprint(api.BLUEPRINT)
    app.register_blueprint(invoices.BLUEPRINT, url_prefix="/invoices")
    app.register_blueprint(oauth_bp, url_prefix="/login")
    _register_admin_views()

    ext.csrf.exempt(api.BLUEPRINT)  # Protected with Auth header already

    @app.route("/")
    def index():
        return redirect(url_for("admin.index"))

    @app.route("/logout")
    def logout():
        """Log out the user."""
        session["user_email"] = None
        return redirect(url_for("index"))


def _register_admin_views():
    # Base data views
    ext.admin.add_view(admin.ApplicationView(Application, ext.db.session))
    ext.admin.add_view(admin.ApplicationVersionView(ApplicationVersion, ext.db.session))
    ext.admin.add_view(admin.ApplicationLimitationsView(ApplicationLimitations, ext.db.session))
    ext.admin.add_view(admin.BedView(Bed, ext.db.session))
    ext.admin.add_view(admin.BedVersionView(BedVersion, ext.db.session))
    ext.admin.add_view(admin.CustomerView(Customer, ext.db.session))
    ext.admin.add_view(admin.CollaborationView(Collaboration, ext.db.session))
    ext.admin.add_view(admin.OrganismView(Organism, ext.db.session))
    ext.admin.add_view(admin.PanelView(Panel, ext.db.session))
    ext.admin.add_view(admin.UserView(User, ext.db.session))
    ext.admin.add_view(
        admin.SampleLaneSequencingMetricsView(SampleLaneSequencingMetrics, ext.db.session)
    )

    # Business data views
    ext.admin.add_view(admin.FamilyView(Family, ext.db.session))
    ext.admin.add_view(admin.FamilySampleView(FamilySample, ext.db.session))
    ext.admin.add_view(admin.SampleView(Sample, ext.db.session))
    ext.admin.add_view(admin.PoolView(Pool, ext.db.session))
    ext.admin.add_view(admin.FlowcellView(Flowcell, ext.db.session))
    ext.admin.add_view(admin.AnalysisView(Analysis, ext.db.session))
    ext.admin.add_view(admin.DeliveryView(Delivery, ext.db.session))
    ext.admin.add_view(admin.InvoiceView(Invoice, ext.db.session))


def _register_teardowns(app: Flask):
    """Register teardown functions."""

    @app.teardown_appcontext
    def remove_database_session(exception=None):
        """
        Remove the database session to ensure database resources are
        released when a request has been processed.
        """
        scoped_session_registry: Optional[scoped_session] = get_scoped_session_registry()
        if scoped_session_registry:
            scoped_session_registry.remove()



"""Interactions with the gisaid cli upload_results_to_gisaid"""
import logging
import re
import tempfile
from pathlib import Path
from typing import Optional

import pandas as pd
from housekeeper.store.models import File

from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.apps.lims import LimsAPI
from cg.constants.constants import SARS_COV_REGEX, FileFormat
from cg.exc import HousekeeperFileMissingError
from cg.io.controller import ReadFile, WriteFile
from cg.models.cg_config import CGConfig
from cg.store import Store
from cg.store.models import Sample
from cg.utils import Process

from .constants import HEADERS
from .models import GisaidAccession, GisaidSample

LOG = logging.getLogger(__name__)

UPLOADED_REGEX_MATCH = r"\[\"([A-Za-z0-9_]+)\"\]\}$"


class GisaidAPI:
    """Interface with Gisaid cli uppload"""

    def __init__(self, config: CGConfig):
        self.housekeeper_api: HousekeeperAPI = config.housekeeper_api
        self.lims_api: LimsAPI = config.lims_api
        self.status_db: Store = config.status_db
        self.gisaid_submitter: str = config.gisaid.submitter
        self.upload_password: str = config.gisaid.upload_password
        self.upload_cid: str = config.gisaid.upload_cid
        self.gisaid_binary: str = config.gisaid.binary_path
        self.gisaid_log_dir: str = config.gisaid.log_dir
        self.log_watch: str = config.gisaid.logwatch_email
        self.email_base_settings = config.email_base_settings
        self.mutant_root_dir = Path(config.mutant.root)

        self.process = Process(binary=self.gisaid_binary)

    def get_completion_file_from_hk(self, case_id: str) -> File:
        """Find completon file in Housekeeper and return it"""

        completion_file: Optional[File] = self.housekeeper_api.get_file_from_latest_version(
            bundle_name=case_id, tags=["komplettering"]
        )
        if not completion_file:
            msg = f"completion file missing for bundle {case_id}"
            raise HousekeeperFileMissingError(message=msg)
        return completion_file

    def get_completion_dataframe(self, completion_file: File) -> pd.DataFrame:
        """Read completion file in to dataframe, drop duplicates, and return the dataframe"""
        completion_df = pd.read_csv(completion_file.full_path, index_col=None, header=0)
        completion_df.drop_duplicates(inplace=True)
        completion_df = completion_df[completion_df["provnummer"].str.contains(SARS_COV_REGEX)]
        return completion_df

    def get_gisaid_sample_list(self, case_id: str) -> list[Sample]:
        """Get list of Sample objects eligeble for upload.
        The criteria is that the sample reached 20x coverage for >95% bases.
        The sample will be included in completion file."""

        completion_file = self.get_completion_file_from_hk(case_id=case_id)
        completion_df = self.get_completion_dataframe(completion_file=completion_file)
        sample_names = list(completion_df["provnummer"].unique())
        return [self.status_db.get_sample_by_name(name=sample_name) for sample_name in sample_names]

    def get_gisaid_fasta_path(self, case_id: str) -> Path:
        """Get path to gisaid fasta"""
        return Path(self.mutant_root_dir, case_id, "results", f"{case_id}.fasta")

    def get_gisaid_csv_path(self, case_id: str) -> Path:
        """Get path to gisaid csv"""
        return Path(self.mutant_root_dir, case_id, "results", f"{case_id}.csv")

    def get_gisaid_samples(self, case_id: str) -> list[GisaidSample]:
        """Get list of Gisaid sample objects."""

        samples: list[Sample] = self.get_gisaid_sample_list(case_id=case_id)
        gisaid_samples = []
        for sample in samples:
            sample_id: str = sample.internal_id
            LOG.info(f"Creating GisaidSample for {sample_id}")
            gisaid_sample = GisaidSample(
                case_id=case_id,
                cg_lims_id=sample_id,
                covv_subm_sample_id=sample.name,
                submitter=self.gisaid_submitter,
                fn=f"{case_id}.fasta",
                covv_collection_date=str(
                    self.lims_api.get_sample_attribute(lims_id=sample_id, key="collection_date")
                ),
                region=self.lims_api.get_sample_attribute(lims_id=sample_id, key="region"),
                region_code=self.lims_api.get_sample_attribute(
                    lims_id=sample_id, key="region_code"
                ),
                covv_orig_lab=self.lims_api.get_sample_attribute(
                    lims_id=sample_id, key="original_lab"
                ),
                covv_orig_lab_addr=self.lims_api.get_sample_attribute(
                    lims_id=sample_id, key="original_lab_address"
                ),
            )
            gisaid_samples.append(gisaid_sample)
        return gisaid_samples

    def create_gisaid_fasta(self, gisaid_samples: list[GisaidSample], case_id: str) -> None:
        """Writing a new fasta with headers adjusted for gisaid upload_results_to_gisaid"""

        gisaid_fasta_file = self.housekeeper_api.get_file_from_latest_version(
            bundle_name=case_id, tags=["gisaid-fasta", case_id]
        )
        if gisaid_fasta_file:
            gisaid_fasta_path = gisaid_fasta_file.full_path
        else:
            gisaid_fasta_path: Path = self.get_gisaid_fasta_path(case_id=case_id)

        fasta_lines: list[str] = []

        for sample in gisaid_samples:
            fasta_file: File = self.housekeeper_api.get_file_from_latest_version(
                bundle_name=case_id, tags=[sample.cg_lims_id, "consensus-sample"]
            )
            if not fasta_file:
                raise HousekeeperFileMissingError(
                    message=f"No fasta file found for sample {sample.cg_lims_id}"
                )
            with open(str(fasta_file.full_path)) as handle:
                for line in handle.readlines():
                    if line[0] == ">":
                        fasta_lines.append(f">{sample.covv_virus_name}\n")
                    else:
                        fasta_lines.append(line)

        with open(gisaid_fasta_path, "w") as write_file_obj:
            write_file_obj.writelines(fasta_lines)

        if gisaid_fasta_file:
            return

        self.housekeeper_api.add_and_include_file_to_latest_version(
            bundle_name=case_id, file=gisaid_fasta_path, tags=["gisaid-fasta", case_id]
        )

    def create_gisaid_csv(self, gisaid_samples: list[GisaidSample], case_id: str) -> None:
        """Create csv file for gisaid upload"""
        samples_df = pd.DataFrame(
            data=[gisaid_sample.model_dump() for gisaid_sample in gisaid_samples],
            columns=HEADERS,
        )

        gisaid_csv_file = self.housekeeper_api.get_file_from_latest_version(
            bundle_name=case_id, tags=["gisaid-csv", case_id]
        )
        if gisaid_csv_file:
            LOG.info(f"GISAID CSV for case {case_id} exists, will be replaced")
            gisaid_csv_path = gisaid_csv_file.full_path
        else:
            gisaid_csv_path = self.get_gisaid_csv_path(case_id=case_id)
        samples_df.to_csv(gisaid_csv_path, sep=",", index=False)

        if gisaid_csv_file:
            return

        self.housekeeper_api.add_and_include_file_to_latest_version(
            bundle_name=case_id, file=gisaid_csv_path, tags=["gisaid-csv", case_id]
        )

    def create_gisaid_log_file(self, case_id: str) -> None:
        """Path for gisaid bundle log"""
        gisaid_log_file = self.housekeeper_api.get_files(
            bundle=case_id, tags=["gisaid-log", case_id]
        ).first()
        if gisaid_log_file:
            LOG.info("GISAID log exists in case bundle in Housekeeper")
            return

        log_file_path = Path(self.gisaid_log_dir, case_id).with_suffix(".log")
        if not log_file_path.parent.exists():
            raise ValueError(f"Gisaid log dir: {self.gisaid_log_dir} doesnt exist")
        if not log_file_path.exists():
            log_file_path.touch()
        self.housekeeper_api.add_and_include_file_to_latest_version(
            bundle_name=case_id, file=log_file_path, tags=["gisaid-log", case_id]
        )

    def create_gisaid_files_in_housekeeper(self, case_id: str) -> None:
        """Create all gisaid files in Housekeeper, if needed."""
        gisaid_samples = self.get_gisaid_samples(case_id=case_id)
        self.create_gisaid_csv(gisaid_samples=gisaid_samples, case_id=case_id)
        self.create_gisaid_fasta(gisaid_samples=gisaid_samples, case_id=case_id)
        self.create_gisaid_log_file(case_id=case_id)

    def authenticate_gisaid(self):
        load_call: list = [
            "CoV",
            "authenticate",
            "--cid",
            self.upload_cid,
            "--user",
            self.gisaid_submitter,
            "--pass",
            self.upload_password,
        ]
        self.process.run_command(parameters=load_call)

    def upload_results_to_gisaid(self, case_id: str) -> None:
        """Load batch data to GISAID using the gisiad cli."""

        temp_log_file = tempfile.NamedTemporaryFile(
            dir=self.gisaid_log_dir, mode="w+", delete=False
        )
        gisaid_csv_path = self.housekeeper_api.get_file_from_latest_version(
            bundle_name=case_id, tags=["gisaid-csv", case_id]
        ).full_path

        gisaid_fasta_path = self.housekeeper_api.get_file_from_latest_version(
            bundle_name=case_id, tags=["gisaid-fasta", case_id]
        ).full_path

        gisaid_log_path = (
            self.housekeeper_api.get_files(bundle=case_id, tags=["gisaid-log", case_id])
            .first()
            .full_path
        )

        self.authenticate_gisaid()
        load_call: list = [
            "--logfile",
            temp_log_file.name,
            "CoV",
            "upload",
            "--csv",
            gisaid_csv_path,
            "--fasta",
            gisaid_fasta_path,
        ]
        self.process.run_command(parameters=load_call)
        self.append_log(temp_log=Path(temp_log_file.name), gisaid_log=Path(gisaid_log_path))
        temp_log_file.close()
        if self.process.stderr:
            LOG.info(f"gisaid stderr:\n{self.process.stderr}")
        if self.process.stdout:
            LOG.info(f"gisaid stdout:\n{self.process.stdout}")

    def append_log(self, temp_log: Path, gisaid_log: Path) -> None:
        """Appends temp log to gisaid log and delete temp file"""
        new_log_data = ReadFile.get_content_from_file(
            file_format=FileFormat.JSON, file_path=temp_log.absolute()
        )
        if gisaid_log.stat().st_size != 0:
            old_log_data = ReadFile.get_content_from_file(
                file_format=FileFormat.JSON, file_path=gisaid_log.absolute()
            )
            new_log_data.extend(old_log_data)

        WriteFile.write_file_from_content(
            content=new_log_data, file_format=FileFormat.JSON, file_path=gisaid_log.absolute()
        )
        temp_log.unlink()

    def get_accession_numbers(self, case_id: str) -> dict[str, str]:
        """Parse accession numbers and sample ids from log file"""

        LOG.info("Parsing accession numbers from log file")
        accession_numbers = {}
        log_file = Path(
            self.housekeeper_api.get_files(bundle=case_id, tags=["gisaid-log", case_id])
            .first()
            .full_path
        )
        if log_file.stat().st_size != 0:
            log_data = ReadFile.get_content_from_file(
                file_format=FileFormat.JSON, file_path=log_file.absolute()
            )
            for log in log_data:
                if log.get("code") == "epi_isl_id":
                    log_message = log.get("msg")
                elif log.get("code") == "validation_error" and "existing_ids" in log.get("msg"):
                    log_message = (
                        f'{log.get("msg").split(";")[0]}; '
                        f'{re.findall(UPLOADED_REGEX_MATCH, log.get("msg"))[0]}'
                    )
                else:
                    continue
                accession_obj = GisaidAccession(log_message=log_message)
                accession_numbers[accession_obj.sample_id] = accession_obj.accession_nr
        return accession_numbers

    def update_completion_file(self, case_id: str) -> None:
        """Update completion file with accession numbers"""
        completion_file = self.get_completion_file_from_hk(case_id=case_id)
        accession_dict = self.get_accession_numbers(case_id=case_id)
        completion_df = self.get_completion_dataframe(completion_file=completion_file)
        completion_df["GISAID_accession"] = completion_df["provnummer"].apply(
            lambda x: accession_dict[x]
        )
        completion_df.to_csv(
            completion_file.full_path,
            sep=",",
            index=False,
        )

    def upload(self, case_id: str) -> None:
        """Uploading results to gisaid and saving the accession numbers in completion file"""

        completion_file = self.get_completion_file_from_hk(case_id=case_id)
        completion_df = self.get_completion_dataframe(completion_file=completion_file)
        if len(completion_df["GISAID_accession"].dropna()) == len(completion_df["provnummer"]):
            LOG.info("All samples already uploaded")
            return

        self.create_gisaid_files_in_housekeeper(case_id=case_id)
        self.upload_results_to_gisaid(case_id=case_id)
        self.update_completion_file(case_id=case_id)



from typing import Optional

from pydantic import BaseModel, Field


class ReadMetric(BaseModel):
    """Metrics for a read for a sample on a tile in a lane."""

    read_yield: int = Field(..., alias="Yield", ge=0)
    read_yield_q30: int = Field(..., alias="YieldQ30", ge=0)
    read_quality_score_sum: int = Field(..., alias="QualityScoreSum", ge=0)


class TileReads(BaseModel):
    """Metrics for a sample on a tile in a lane."""

    sample_id: str = Field(..., alias="SampleId", min_length=1)
    tile_sample_reads: int = Field(..., alias="NumberReads", ge=0)
    tile_sample_yield: int = Field(..., alias="Yield", ge=0)
    tile_sample_read_metrics: list[ReadMetric] = Field(..., alias="ReadMetrics")


class UndeterminedTileReads(BaseModel):
    """Metrics for reads without a sample id within a tile of a lane."""

    tile_total_reads: int = Field(..., alias="NumberReads", ge=0)
    tile_total_yield: int = Field(..., alias="Yield", ge=0)
    tile_read_metrics: list[ReadMetric] = Field(..., alias="ReadMetrics")


class ConversionResult(BaseModel):
    """Demultiplexing results and undetermined reads for a tile within a lane."""

    lane: int = Field(..., alias="LaneNumber", gt=0)
    tile_reads: list[TileReads] = Field(..., alias="DemuxResults")
    tile_undetermined_reads: Optional[UndeterminedTileReads] = Field(None, alias="Undetermined")


class SampleLaneTileMetrics(BaseModel):
    """Metrics for samples on a tile in a lane on a flow cell from a bcl2fastq run."""

    flow_cell_name: str = Field(..., alias="Flowcell", min_length=1)
    conversion_results: list[ConversionResult] = Field(..., alias="ConversionResults")


class SampleLaneMetrics(BaseModel):
    """Metrics for a sample in a lane from a bcl2fastq run."""

    flow_cell_name: str
    lane: int
    sample_id: str
    total_reads: int
    total_yield: int
    total_yield_q30: int
    total_quality_score: int
